//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com

#include <sstream>
#include "data.h"
#include "data/row_data.h"
#include "data/row_data.cpp"

const std::string HEX_CHARS="0123456789abcdef";

u16 dehex(std::string s)
{
  u16 res(0);
  u8 size(s.size());
  for (u8 i=0; i < size; ++i)
  {
    if (HEX_CHARS.find(s[i]))
    {
      res += (HEX_CHARS.find(s[i])) << (4 * (size - i - 1));
    }
  }
  return res;
}

//--------------------------------------------------------------------

u16
Reading::operator[](u8 index) const
{
  if (index < size())
  {
    std::string double_char = m_strEncoded.substr((index * 2), 2);
    if (double_char == "..") { return SEPARATOR; }
    else if (double_char == "??") { return QMARK; }
    else if (HEX_CHARS.find(double_char[0]) != std::string::npos
	     && HEX_CHARS.find(double_char[1]) != std::string::npos)
    {
      return dehex(double_char);
    }
    else { return BAD_CHAR; }
  }
  else { return BAD_CHAR; }
}

//--------------------------------------------------------------------

bool
Reading::compare(const Reading &x, const Reading &y)
{
  bool equal(true);
  bool res(false);
  u16 x_char(BAD_CHAR);
  u16 y_char(BAD_CHAR);
  u8 x_ind(0);
  u8 y_ind(0);
  while (equal && x_ind < x.size() && y_ind < y.size())
  {
    x_char = x[x_ind];
    y_char = y[y_ind];
    if (x_char == SEPARATOR || x_char == BAD_CHAR) { ++x_ind; }
    else if (y_char == SEPARATOR || y_char == BAD_CHAR) { ++y_ind; }
    else
    {
      if (x_char >= 128) { x_char -= 96; }
      if (y_char >= 128) { y_char -= 96; }
      equal = (x_char == y_char);
      res = (x_char < y_char);
      x_ind += 1;
      y_ind += 1;
    }
  }
  if (equal)
  {
    res = (x.size() < y.size());
  }
  return res;
}

//--------------------------------------------------------------------

// Encoding format:
// 4 chars: unicode value in hex, ex. "40e0"
// 1 char : grade
// all other chars: the reading
ContentRow::ContentRow(std::string s)
  : m_nUnicode(0), m_nGrade(0), m_reading("")
{
  static u16 calls(0);
  if (s.size() > 5)
  {
    calls += 1;
    m_nUnicode = dehex(s.substr(0, 4));
    m_nGrade = dehex(s.substr(4, 1));
    m_reading = Reading(s.substr(5));
  }
  else { m_nGrade = s.size() + 20; }
}

//--------------------------------------------------------------------

//--------------------------------------------------------------------

bool
readingCompare(const ContentRow &x, const ContentRow &y)
{
  return Reading::compare(x.reading(), y.reading());
}

//--------------------------------------------------------------------

bool
rowCompare(const u16 &x, const u16 &y)
{
  // change ContentData so size is cached.
  u16 size = GameData::content.size();
  if (x < size && y < size)
  { 
    return readingCompare(GameData::content.getRow(x),
			  GameData::content.getRow(y));
  }
  else { return false; }
}

//--------------------------------------------------------------------

ContentData::ContentData()
  : m_nSize(0)
{
  m_vecContent.clear();

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE1; ++i)
  { m_vecContent.push_back(ContentRow(data_grade1[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE2; ++i)
  { m_vecContent.push_back(ContentRow(data_grade2[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE3; ++i)
  { m_vecContent.push_back(ContentRow(data_grade3[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE4; ++i)
  { m_vecContent.push_back(ContentRow(data_grade4[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE5; ++i)
  { m_vecContent.push_back(ContentRow(data_grade5[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE6; ++i)
  { m_vecContent.push_back(ContentRow(data_grade6[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE7; ++i)
  { m_vecContent.push_back(ContentRow(data_grade7[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE8; ++i)
  { m_vecContent.push_back(ContentRow(data_grade8[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADE9; ++i)
  { m_vecContent.push_back(ContentRow(data_grade9[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADEA; ++i)
  { m_vecContent.push_back(ContentRow(data_gradeA[i])); }

  m_vecFirstIndices.push_back(m_vecContent.size());
  for (u16 i=0; i < SIZE_GRADEB; ++i)
  { m_vecContent.push_back(ContentRow(data_gradeB[i])); }

  m_nSize = m_vecContent.size();
}

//--------------------------------------------------------------------

ContentRow
ContentData::random() const
{
  return m_vecContent[PA_RandMax(m_vecContent.size()-1)];
}

//--------------------------------------------------------------------

ReadingVec
ContentData::getReadings(u16 row_ind)
{
  u16 i(row_ind);
  ReadingVec res;
  while (i>0 && m_vecContent[i].unicode() == m_vecContent[row_ind].unicode())
  {
    i -= 1;
  }
  if (m_vecContent[i].unicode() != m_vecContent[row_ind].unicode()) { i += 1; }

  while (i < m_vecContent.size()-1
	 && m_vecContent[i].unicode() == m_vecContent[row_ind].unicode())
  {
    res.push_back(m_vecContent[i].reading());
    i += 1;
  }
  return res;
}

//--------------------------------------------------------------------

std::pair<u16, u16>
ContentData::getCharRange(u16 row_ind)
{
  u16 i(row_ind);
  std::pair<u16, u16> res;
  while (i>0 && m_vecContent[i].unicode() == m_vecContent[row_ind].unicode())
  {
    i -= 1;
  }
  if (m_vecContent[i].unicode() != m_vecContent[row_ind].unicode()) { i += 1; }
  res.first = i;
  res.second = 1;
  while (i < m_vecContent.size()-1
	 && m_vecContent[i].unicode() == m_vecContent[row_ind].unicode())
  {
    i += 1;
  }
  res.second = i;
  return res;
}

//--------------------------------------------------------------------

PlayerRecord::PlayerRecord()
  : m_nMasteredCutoff(66), m_queueNew(12), m_queueWrong(12)
{
  for (u8 i=0; i < GameData::content.numGrades(); ++i)
  {
    GradeLevelSet gls(GameData::content.firstIndex(i),
		      GameData::content.size(i));
    m_vecGradeLevelSets.push_back(gls);
  }
}

//--------------------------------------------------------------------

u16
PlayerRecord::randomByStatus(RowState s, u16 limit)
{
  u16 res(ContentData::BAD_ROW);
  u16 num_rows(0);
  u8 max_grade(0);

  if (limit)
  {
    for (u8 i=0; i < m_vecGradeLevelSets.size() && num_rows < limit; ++i)
    {
      num_rows += m_vecGradeLevelSets[i].sizeByStatus(s);
      max_grade = i+1;
    }
    if (num_rows > limit) { num_rows = limit; }
  }
  else
  {
    for (u8 i=0; i < m_vecGradeLevelSets.size(); ++i)
    {
      num_rows += m_vecGradeLevelSets[i].sizeByStatus(s);
    }
    max_grade = m_vecGradeLevelSets.size();
  }

  if (num_rows == 0) { return ContentData::BAD_ROW; }

  u16 row = PA_RandMax(num_rows - 1);
  for (u8 i=0; i < max_grade && res == ContentData::BAD_ROW; ++i)
  {
    if (row < m_vecGradeLevelSets[i].sizeByStatus(s))
    {
      row = PA_RandMax(m_vecGradeLevelSets[i].sizeByStatus(s) - 1);
      res = m_vecGradeLevelSets[i].nthByStatus(s, row);
    }
    else { row -= m_vecGradeLevelSets[i].sizeByStatus(s); }
  }

  return res;
}

//--------------------------------------------------------------------

u16
PlayerRecord::pick(const std::vector<u16> &excluded)
{
  // check new queue.
  if (m_queueNew.small())
  {
    for (u8 i=0; i < m_queueNew.sizeFactor(); ++i)
    {
      u16 row = randomByStatus(New, 20);
      if (row != ContentData::BAD_ROW) { m_queueNew.add(row); }
    }
  }

  // check wrong queue.
  if (m_queueWrong.small())
  {
    for (u8 i=0; i < m_queueWrong.sizeFactor(); ++i)
    {
      u16 row = randomByStatus(Wrong);
      if (row != ContentData::BAD_ROW) { m_queueWrong.add(row); }
    }
  }

  // Body of whichSet, chooses the state of the pick.
  RowState state = Mastered;
  if (PA_RandMax(100) >= m_nMasteredCutoff)
  {
    if (m_nNewLocks) { state = Wrong; }
    else
    {
      if (PA_RandMax(1) && !m_queueWrong.empty()) { state = Wrong; }
      else { state = New;}
    }
  }

  u16 res(ContentData::BAD_ROW);
  if (state == New) { res = m_queueNew.randomExclude(excluded); }
  else if (state == Wrong) { res = m_queueWrong.randomExclude(excluded); }

  u8 num_tries(0);
  while (res == ContentData::BAD_ROW && num_tries < 3)
  {
    res = randomByStatus(state);
    if (find(excluded.begin(), excluded.end(), res) != excluded.end())
    {
      res = ContentData::BAD_ROW;
    }
    num_tries += 1;
  }
  if (res == ContentData::BAD_ROW) { res = randomByStatus(Wrong); }
  if (res == ContentData::BAD_ROW) { res = randomByStatus(New, 30); }
  if (res == ContentData::BAD_ROW) { res = randomByStatus(Mastered); }

  return res;
}

//--------------------------------------------------------------------

void
PlayerRecord::lowerMasteredCutoff()
{
  if (m_nMasteredCutoff > 10) { m_nMasteredCutoff -= 1; }
}

//--------------------------------------------------------------------

void
PlayerRecord::raiseMasteredCutoff()
{
  if (m_nMasteredCutoff < 80)
  {
    m_nMasteredCutoff += (80 - m_nMasteredCutoff) / 2;
  }
}

//--------------------------------------------------------------------

PlayerRecord::EditResult
PlayerRecord::recordCorrect(u16 row_ind)
{
  EditResult result(NoAction);
  for (std::vector<GradeLevelSet>::iterator iter=m_vecGradeLevelSets.begin();
       iter != m_vecGradeLevelSets.end() && result == NoAction;
       ++iter)
  {
    result = iter->recordCorrect(row_ind);
  }
  return result;
}

//--------------------------------------------------------------------

PlayerRecord::EditResult
PlayerRecord::recordIncorrect(u16 row_ind)
{
  EditResult result(NoAction);
  for (std::vector<GradeLevelSet>::iterator iter=m_vecGradeLevelSets.begin();
       iter != m_vecGradeLevelSets.end() && result == NoAction;
       ++iter)
  {
    result = iter->recordIncorrect(row_ind);
  }
  return result;
}

//--------------------------------------------------------------------

u8
PlayerRecord::currentGrade() const
{
  if (m_vecGradeLevelSets.empty()) { return 0; }
  u8 res = m_vecGradeLevelSets.size() - 1;
  bool found=false;
  for (u8 i=0; i < m_vecGradeLevelSets.size() && !found; ++i)
  {
    if (m_vecGradeLevelSets[i].sizeByStatus(New)
	+ m_vecGradeLevelSets[i].sizeByStatus(Wrong)
	> 20)
    {
      found = true;
      res = i;
    }
  }
  return res;
}

//--------------------------------------------------------------------

u16
PlayerRecord::sizeByStatus(u8 grade, RowState state) const
{
  if (grade < m_vecGradeLevelSets.size())
  {
    return m_vecGradeLevelSets[grade].sizeByStatus(state);
  }
  else { return 0; }
}

//--------------------------------------------------------------------

u8
PlayerRecord::gradeCompletionPercentage(u8 i) const
{
  if (i < m_vecGradeLevelSets.size())
  {
    return m_vecGradeLevelSets[i].completionPercentage();
  }
  else { return 0; }
}

//--------------------------------------------------------------------

void
PlayerRecord::write(DataFile &file) const
{
  for (unsigned int i=0; i < m_vecGradeLevelSets.size(); ++i)
  {
    std::stringstream ss;
    ss << "grade" << i + 1;
    file.addVec(ss.str(), m_vecGradeLevelSets[i].vectorize());
  }
}

//--------------------------------------------------------------------

void
PlayerRecord::read(const DataFile &file)
{
  for (unsigned int i=0; i < m_vecGradeLevelSets.size(); ++i)
  {
    std::stringstream ss;
    ss << "grade" << i + 1;
    // if not in the file, make a new gls.
    std::vector<u32> grade_vec(file.lookupVec(ss.str()));
    if (grade_vec.empty())
    {
      m_vecGradeLevelSets[i] = GradeLevelSet(GameData::content.firstIndex(i),
					     GameData::content.size(i));
    }
    else { m_vecGradeLevelSets[i].devectorize(grade_vec); }
  }
}

//--------------------------------------------------------------------

std::vector<u16>
PlayerRecord::getReadingSortedRows()
{
  std::vector<u16> res;
  for (std::vector<GradeLevelSet>::const_iterator iter = m_vecGradeLevelSets.begin();
       iter != m_vecGradeLevelSets.end();
       ++iter)
  {
    std::set<u16> row_set = iter->masteredRows();
    res.insert(res.end(), row_set.begin(), row_set.end());
  }
  std::sort(res.begin(), res.end(), rowCompare);
  return res;
}

//--------------------------------------------------------------------


PlayerRecord::RowState
PlayerRecord::state(u16 row_ind) const
{
  RowState res(New);
  for (std::vector<GradeLevelSet>::const_iterator iter = m_vecGradeLevelSets.begin();
       iter != m_vecGradeLevelSets.end();
       ++iter)
  {
    if (iter->contains(row_ind)) { res = iter->state(row_ind); }
  }
  return res;
}

//--------------------------------------------------------------------

void
PlayerRecord::debug() const
{
  for (u16 i=0; i < 10; ++i)
  {
    PA_OutputText(1, 0, 13 + i, "new item #%d = %d", i,
		  m_vecGradeLevelSets[0].nthByStatus(New, i));
  }

  m_queueNew.debug();
}

//--------------------------------------------------------------------

PlayerRecord::GradeLevelSet::GradeLevelSet(u16 offset, u16 expected_size)
  : m_nOffset(offset), m_nSize(expected_size)
{
  for (u16 i=offset; i < offset + expected_size; ++i)
  {
    m_mapSets[New].insert(i);
  }
  m_mapSetSizes[New] = m_nSize;

  // init count vecs.
  // need an extra index.
  m_mapCounts[Mastered] = CountVec((m_nSize / 32) + 1, 0);
  m_mapCounts[Wrong] = CountVec((m_nSize / 32) + 1, 0);
  m_mapCounts[Hidden] = CountVec((m_nSize / 32) + 1, 0);
  m_mapCounts[New] = CountVec((m_nSize / 32) + 1, 32);
  m_mapCounts[New][m_mapCounts[New].size() - 1] = m_nSize % 32;
}

//--------------------------------------------------------------------

PlayerRecord::EditResult
PlayerRecord::GradeLevelSet::recordCorrect(u16 row_ind)
{
  if (contains(row_ind))
  {
    RowState s(state(row_ind));
    if (s == Wrong || s == New)
    {
      RowState source(s);
      RowState destination(New);
      if (s == Wrong) { destination = New; }
      else if (s == New) { destination = Mastered; }

      remove(source, row_ind);
      insert(destination, row_ind);

      return Master;
    }
    else { return Recorded; }
  }
  else { return NoAction; }
}

//--------------------------------------------------------------------

PlayerRecord::EditResult
PlayerRecord::GradeLevelSet::recordIncorrect(u16 row_ind)
{
  if (contains(row_ind))
  {
    RowState s(state(row_ind));
    if (s == New || s == Mastered)
    {
      remove(s, row_ind);
      insert(Wrong, row_ind);
      return Unmaster;
    }
    else { return Recorded; }
  }
  else { return NoAction; }
}

//--------------------------------------------------------------------

PlayerRecord::RowState
PlayerRecord::GradeLevelSet::state(u16 row_ind) const
{
  RowState res(Hidden);
  for (std::map<RowState, RowSet>::const_iterator iter=m_mapSets.begin();
       iter != m_mapSets.end();
       ++iter)
  {
    if (iter->second.find(row_ind) != iter->second.end())
    {
      res = iter->first;
    }
  }
  return res;
}

//--------------------------------------------------------------------

void
PlayerRecord::GradeLevelSet::insert(RowState s, u16 row_ind)
{
  // .second because insert() returns pair<iterator, bool>
  if (m_mapSets[s].insert(row_ind).second)
  {
    m_mapCounts[s][(row_ind - m_nOffset) / 32] += 1;
    m_mapSetSizes[s] += 1;
  }
}

//--------------------------------------------------------------------

void
PlayerRecord::GradeLevelSet::remove(RowState s, u16 row_ind)
{
  if (m_mapSets[s].erase(row_ind))
  {
    m_mapCounts[s][(row_ind - m_nOffset) / 32] -= 1;
    m_mapSetSizes[s] -= 1;
  }
}

//--------------------------------------------------------------------

u16
PlayerRecord::GradeLevelSet::sizeByStatus(RowState s) const
{
  std::map<RowState, RowSet>::const_iterator iter = m_mapSets.find(s);
  if (iter != m_mapSets.end()) { return iter->second.size(); }
  else { return 0; }
}

//--------------------------------------------------------------------

// If this function just iterated over all the row indices,
// calling state() on each and keeping count,
// the function would be O(n log n), making O(n) calls to state().
// Now it makes at most 32 calls to state(), an improvement
// when n is in the hundreds.

u16
PlayerRecord::GradeLevelSet::nthByStatus(RowState s, u16 n) const
{
  u16 res(ContentData::BAD_ROW);

  std::map<RowState, CountVec>
    ::const_iterator count_map_iter(m_mapCounts.find(s));

  if (count_map_iter == m_mapCounts.end()) { return 0; }

  for (u8 i=0;
       i < count_map_iter->second.size() && res == ContentData::BAD_ROW;
       ++i)
  {
    if (n < count_map_iter->second[i])
    {
      u8 local_count(0);
      for (u16 row=m_nOffset + (32 * i);
	   row < m_nOffset + (32 * (i+1)) && res == ContentData::BAD_ROW;
	   ++row)
      {
	if (state(row) == s)
	{
	  if (local_count == n) { res = row; }
	  else { local_count += 1; }
	}
      }
    }
    else { n -= count_map_iter->second[i]; }
  }

  if (res == ContentData::BAD_ROW)
  {
    PA_OutputText(1, 0, 21, "nthByStatus failed for state %d row %d", s, n);
  }

  return res;
}

//--------------------------------------------------------------------

u8
PlayerRecord::GradeLevelSet::completionPercentage() const
{
  if (m_nSize) { return (sizeByStatus(Mastered) * 100) / m_nSize; }
  else { return 0; }
}

//--------------------------------------------------------------------

PlayerRecord::GradeLevelSet::RowSet
PlayerRecord::GradeLevelSet::masteredRows() const
{
  std::map<RowState, RowSet>::const_iterator map_iter(m_mapSets.find(Mastered));

  if (map_iter == m_mapSets.end()) { return RowSet(); }
  else { return map_iter->second; }
}

//--------------------------------------------------------------------

std::vector<u32>
PlayerRecord::GradeLevelSet::vectorize() const
{
  std::vector<u32> res;
  for (u8 i=0; i < (m_nSize / 32) + 1; ++i)
  {
    u32 val(0);
    for (u8 j=0; j < 32; ++j)
    {
      u16 row = m_nOffset + j + (32 * i);
      if (row < m_nOffset + m_nSize && state(row) == Mastered)
      {
	val += (1 << j);
      }
    }
    res.push_back(val);
  }
  return res;
}

//--------------------------------------------------------------------

void
PlayerRecord::GradeLevelSet::devectorize(const std::vector<u32> &v)
{
  for (u8 i=0; i < (m_nSize / 32) + 1 && i < v.size(); ++i)
  {
    for (u8 j=0; j < 32; ++j)
    {
      u16 row = m_nOffset + j + (32 * i);
      if (row < m_nOffset + m_nSize)
      {
	RowState current_state = state(row);
	if (v[i] & (1 << j))
	{
	  remove(current_state, row);
	  insert(Mastered, row);
	}
	else
	{
	  remove(current_state, row);
	  insert(New, row);
	}
      }
    }
  }
}

//--------------------------------------------------------------------

RowQueue::RowQueue(u8 sf)
  : m_nSizeFactor(sf)
{}

//--------------------------------------------------------------------

u16
RowQueue::randomExclude(const std::vector<u16> &excluded)
{
  std::vector<u16> valid;

  for (std::vector<u16>::const_iterator iter=m_vecData.begin();
       iter != m_vecData.end(); ++iter)
  {
    if (find(excluded.begin(), excluded.end(), *iter) == excluded.end())
    {
      valid.push_back(*iter);
    }
  }

  if (!valid.empty())
  {
    u16 res = valid[PA_RandMax(valid.size() - 1)];
    remove(res);
    return res;
  }
  else { return ContentData::BAD_ROW; }
}

//--------------------------------------------------------------------

void
RowQueue::debug() const
{
  std::stringstream ss;
  for (u8 i=0; i < m_vecData.size(); ++i)
  {
    ss << " " << m_vecData[i];
  }
  PA_OutputText(1, 0, 0, const_cast<char*>(ss.str().c_str()));
}

//--------------------------------------------------------------------

ContentData GameData::content;
PlayerRecord GameData::record;
DataFile GameData::file;

//--------------------------------------------------------------------

bool
GameData::readNumberedFile(unsigned int i, PlayerRecord &pr)
{
  DataFile file;
  
  std::stringstream ss;
  ss << "ks/game" << i << ".txt";

  bool res = file.read(ss.str());
  pr.read(file);
  return res;
}

//---------------------------------------------------------------------

bool
GameData::writeNumberedFile(unsigned int i, const PlayerRecord &pr)
{
  DataFile file;
  
  std::stringstream ss;
  ss << "ks/game" << i << ".txt";

  pr.write(file);
  return file.write(ss.str());
}

//---------------------------------------------------------------------

bool
GameData::writeEmptyFile(unsigned int i)
{
  DataFile file;
  
  std::stringstream ss;
  ss << "ks/game" << i << ".txt";

  return file.write(ss.str());
}
