//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com


#ifndef DATA_H
#define DATA_H

#include <PA9.h>
#include <map>
#include <queue>
#include <set>
#include <string>
#include <vector>
#include "file.h"

class Reading
{
 private:
  std::string m_strEncoded;
  bool m_bOnyomi;

 public:
  static const u16 BAD_CHAR=0;
  static const u16 SEPARATOR=1;
  static const u16 QMARK=2;

 public:
 Reading(std::string s)
   : m_strEncoded(s), m_bOnyomi(false)
  {}

  u8 size() const { return u8(m_strEncoded.size() / 2); }
  u16 operator[](u8 index) const;
  bool operator==(const Reading &x) const
  {
    return (this->m_strEncoded == x.m_strEncoded);
  }

  static bool compare(const Reading &x, const Reading &y);
};

typedef std::vector<Reading> ReadingVec;

class ContentRow
{
 private:
  u16 m_nUnicode;
  u8 m_nGrade;
  Reading m_reading;

 private:
 ContentRow(u16 u, u8 g, Reading r)
   : m_nUnicode(u), m_nGrade(g), m_reading(r)
  {}

 public:
  ContentRow(std::string s);
  u16 unicode() const { return m_nUnicode; }
  u8 grade() const { return m_nGrade; }
  Reading reading() const { return m_reading; }
};

bool readingCompare(const ContentRow &x, const ContentRow &y);
bool rowCompare(const u16 &x, const u16 &y);

class ContentData
{
 private:
  std::vector<ContentRow> m_vecContent;
  std::vector<u16> m_vecFirstIndices;
  u16 m_nSize;

 private:
  void addRow(ContentRow r);

 public:
  static const u16 BAD_ROW=-1;

 public:
  ContentData();
  // change to operator[]
  ContentRow getRow(u16 index) const { return m_vecContent[index]; }
  ContentRow random() const;
  u16 size() const { return m_nSize; }
  //  u16 size() const { return m_vecContent.size(); }
  u8 numGrades() const { return m_vecFirstIndices.size(); }
  ReadingVec getReadings(u16 row_ind);

  // getCharRange() returns a range [x, y) that contains rows
  // with the same kanji as row_ind.
  std::pair<u16, u16> getCharRange(u16 row_ind);

  u16 firstIndex(u8 grade) const { return m_vecFirstIndices[grade]; }
  u16 size(u8 grade) const
  {
    if (grade < numGrades() - 1)
    {
      return firstIndex(grade + 1) - firstIndex(grade);
    }
    else { return size() - firstIndex(grade); }
  }
};

class RowQueue
{
 private:
  u8 m_nSizeFactor;
  std::vector<u16> m_vecData;

 public:
  RowQueue(u8 sf);
  void add(u16 elem) { m_vecData.push_back(elem); }
  void remove(u16 elem)
  {
    std::vector<u16>::iterator iter = std::find(m_vecData.begin(),
						m_vecData.end(),
						elem);
    if (iter != m_vecData.end()) { m_vecData.erase(iter); }
  }
  bool big() const { return ((u8)m_vecData.size() > m_nSizeFactor * 2); }
  bool small() const { return ((u8)m_vecData.size() < m_nSizeFactor); }
  bool empty() const { return m_vecData.empty(); }
  u8 size() const { return m_vecData.size(); }
  u8 sizeFactor() const { return m_nSizeFactor; }
  std::vector<u16> data() const { return m_vecData; }
  u16 randomExclude(const std::vector<u16> &excluded);

  void debug() const;
};

// This class combines classes
// GameplayStatus and SaveFile
// from the Python implementation.
class PlayerRecord
{
 public:
  enum RowState { New, Wrong, Mastered, Hidden };
  enum EditResult { NoAction, Recorded, Master, Unmaster };

 private:
  class GradeLevelSet;
  static const u8 NEW_LOCK_PLAYER=1;
  static const u8 NEW_LOCK_GAME=2;

 private:
  u8 m_nMasteredCutoff;  // value in range [0, 100], used in pick()
  u8 m_nNewLocks;
  RowQueue m_queueNew;
  RowQueue m_queueWrong;
  std::vector<GradeLevelSet> m_vecGradeLevelSets;

 private:
  RowState whichSet();
  u16 randomByStatus(RowState s, u16 limit=0);

 public:
  PlayerRecord();
  u16 pick(const std::vector<u16> &excluded);
  void lowerMasteredCutoff();
  void raiseMasteredCutoff();
  EditResult recordCorrect(u16 row_ind);
  EditResult recordIncorrect(u16 row_ind);

  // Based on completion rate.
  u8 currentGrade() const;

  u16 sizeByStatus(u8 grade, RowState state) const;

  // On a scale from 0 to 100.
  u8 gradeCompletionPercentage(u8 i) const;
  u8 totalCompletionPercentage() const;

  void write(DataFile &file) const;
  void read(const DataFile &file);

  std::vector<u16> getReadingSortedRows();
  RowState state(u16 row_ind) const;

  // Just a place to put debugging code.
  void debug() const;
};

typedef PlayerRecord::EditResult RecordEditRes;

// NOTE: The compiler may try to inline GradeLevelSet's
// copy constructor and destructor and fail due to function length.
// In that case, add the following line to CFLAGS in PA_Makefile:
// --param large-function-growth=1000
class PlayerRecord::GradeLevelSet
{
 public:

  typedef std::set<u16> RowSet;

  // StatusIterator class to be implemented later.
  // Iterates both by row and by kanji bidirectionally.

 private:
  static const u8 COUNT_BLOCK_SIZE=32;
  typedef std::vector<u32> CountVec;

 private:
  std::map<RowState, RowSet> m_mapSets;
  std::map<RowState, u16> m_mapSetSizes;
  std::map<RowState, CountVec> m_mapCounts;
  u16 m_nOffset;
  u16 m_nSize;

 private:
  void setBit(bool one, u16 row_ind);
  void insert(RowState s, u16 row_ind);
  void remove(RowState s, u16 row_ind);

 public:
  GradeLevelSet(u16 offset, u16 expected_size);

  RowState state(u16 row_ind) const;

  EditResult recordCorrect(u16 row_ind);
  EditResult recordIncorrect(u16 row_ind);

  u16 sizeByStatus(RowState s) const;
  u16 nthByStatus(RowState s, u16 n) const;
  u8 completionPercentage() const;

  // make private?
  bool contains(u16 row_ind) const
  {
    return (row_ind >= m_nOffset) && (row_ind < m_nOffset + m_nSize);
  }

  RowSet masteredRows() const;

  // For file I/O
  std::vector<u32> vectorize() const;
  void devectorize(const std::vector<u32> &v);
};

namespace GameData
{
  extern ContentData content;
  extern PlayerRecord record;
  extern DataFile file;

  // stringstream doesn't like u8 (maybe?),
  // so make this an unsigned int instead.
  bool readNumberedFile(unsigned int i, PlayerRecord &pr);
  bool writeNumberedFile(unsigned int i, const PlayerRecord &pr);
  bool writeEmptyFile(unsigned int i);
};

#endif
