//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com

#include <cstdlib>
#include "gfx/all_gfx.h"
#include "gameplay.h"
#include "get_chars.h"

//---------------------------------------------------------------------

AnimationState::AnimationState(u8 s, u8 minf, u8 maxf, u8 c, u8 speed)
  : m_nSprite(s), m_nMinFrame(minf), m_nMaxFrame(maxf),
    m_nCycles(c), m_nSpeed(speed), m_bAlive(false),
    m_nCurrentFrame(0), m_nCurrentCycle(0)
{}

//---------------------------------------------------------------------

void
AnimationState::reset()
{
  m_nCurrentFrame = 0;
  m_nCurrentCycle = 0;
}

//---------------------------------------------------------------------

void
AnimationState::update()
{
  if (!m_bAlive)
  { 
    PA_DualSetSpriteAnim(m_nSprite, 0);
    return;
  }

  m_nCurrentFrame = PA_DualGetSpriteAnimFrame(m_nSprite);
  if (m_nCurrentFrame == m_nMaxFrame && m_nCurrentCycle % 2 == 0)
  {
    m_nCurrentCycle += 1;
  }
  else if (m_nCurrentFrame == m_nMinFrame && m_nCurrentCycle % 2)
  {
    m_nCurrentCycle += 1;
  }

  if (m_nCurrentCycle == m_nCycles)
  {
    m_bAlive = false;
    m_nCurrentCycle = 0;
    m_nCycles = 0;
  }
}

//---------------------------------------------------------------------

void
AnimationState::start()
{
  m_bAlive = true;
  if (m_nCycles > m_nCurrentCycle)
  {
    PA_DualStartSpriteAnimEx(m_nSprite, m_nMinFrame, m_nMaxFrame, m_nSpeed,
			     ANIM_UPDOWN, m_nCycles - m_nCurrentCycle);
  }
}

//---------------------------------------------------------------------

void
AnimationState::restart()
{
  if (m_bAlive) { start(); }
}

//---------------------------------------------------------------------

void AnimationState::kill() { m_bAlive = false; }

//---------------------------------------------------------------------

void
AnimationState::display(u8 screen, u8 line)
{
  PA_OutputText(screen, 0, line, "%d %d->%d %d | %d %d %d",
		m_nSprite, m_nMinFrame, m_nMaxFrame, m_nCycles,
		m_nCurrentFrame, m_nCurrentCycle, m_bAlive);
}

//---------------------------------------------------------------------

Meteor::Meteor(s32 a, s32 b, u8 sc, u8 sm, u32 tpm, u8 tid, u16 offset,
	       u8 screen, u8 line)
  : m_nX(a), m_nY(b),
    m_nContent(sc), m_nMeteor(sm),
    m_nTicksPerMove(tpm), m_nTimerID(tid), m_nLastTick(Tick(tid)),
    m_nState(Normal),
    m_nRefreshStartTime(0), m_nRowIndex(0),
    m_nCollisionType(None),
    m_nInitialOffset(offset),
    m_bPaused(false),
    m_nFrame(0),
    m_nAnimState(m_nMeteor, 0, 0, 0, 10),
    m_nOutputScreen(screen),
    m_nOutputLine(line)
{
  s32 half_width = SIZE_CONTENT_SPRITE / 2;
  PA_DualSetSpriteXY(m_nContent,
		     getContentX() - half_width, -SIZE_CONTENT_SPRITE);
  PA_DualSetSpriteXY(m_nMeteor, m_nX, -SIZE_CONTENT_SPRITE);  
  setContentRow(PA_RandMax(GameData::content.size()));
}

//--------------------------------------------------------------------

void
Meteor::setContentRow(u16 row_ind)
{
  m_nRowIndex = row_ind;
  ContentRow row = GameData::content.getRow(row_ind);
  PA_DualUpdateSpriteGfx(m_nContent,
			 (void *)getKanji(row.unicode()));
}

//--------------------------------------------------------------------

void
Meteor::startExplodeAnimation()
{
  m_nCollisionType = None;
  m_nAnimState = AnimationState(m_nMeteor, 2, 6, 1, 10);
  m_nAnimState.start();
  PA_DualSetSpriteXY(m_nContent, getContentX(), -SIZE_CONTENT_SPRITE);
  m_nState = Exploding;
}

//--------------------------------------------------------------------

void
Meteor::startRefreshAnimation()
{
  m_nAnimState = AnimationState(m_nMeteor, 0, 1, 2, 5);
  m_nAnimState.start();
  PA_DualSetSpriteXY(m_nContent, getContentX(), -SIZE_CONTENT_SPRITE);
  m_nRefreshStartTime = Tick(m_nTimerID);
  m_nState = Refresh;
}

//--------------------------------------------------------------------

void
Meteor::startWrongAnimation()
{
  m_nAnimState = AnimationState(m_nMeteor, 7, 8, 4, 10);
  m_nAnimState.start();

  if (m_nCollisionType != RealAnswer)
  {
    PA_DualSetSpriteXY(m_nContent, getContentX(), -SIZE_CONTENT_SPRITE);
    m_nState = StateBadAnswer;
  }
  else { m_nState = StateRealAnswer; }
  m_nCollisionType = None;
}

//--------------------------------------------------------------------

void
Meteor::resetLastTick()
{
  m_nLastTick = Tick(m_nTimerID);
}

//--------------------------------------------------------------------

void
Meteor::move()
{
  if (m_bPaused) { return; }

  m_nAnimState.update();
  if (m_nState != Exploding && m_nState != Paused)
  {
    u32 current_tick = Tick(m_nTimerID);
    u8 dy = (current_tick - m_nLastTick) / m_nTicksPerMove;

    if (dy)
    {
      if (m_nInitialOffset && m_nLastTick)
      {
	if (m_nInitialOffset >= dy) { m_nInitialOffset -= dy; }
	else { m_nInitialOffset = 0; }
      }      
      else if (m_nLastTick) { m_nY += dy; }

      m_nLastTick = current_tick;

      if (m_nY > 384) { m_nY = -32; }

      if (m_nState == Normal || m_nState == StateRealAnswer)
      {
	PA_DualSetSpriteXY(m_nContent, getContentX(), getContentY());
      }

      if (m_nState == Refresh && atRefreshAnimationEnd())
      {
	m_nState = Normal;
      }

      if ((m_nState == StateBadAnswer || m_nState == StateRealAnswer)
	  && !m_nAnimState.alive())
      {
	m_nState = Normal;
      }

      PA_DualSetSpriteXY(m_nMeteor, getMeteorX(), getMeteorY());
    }
  }
  else if (m_nState == Exploding && atExplodeAnimationEnd())
  {
    reset();
    m_nState = Normal;
  }
}

//--------------------------------------------------------------------

void
Meteor::reset()
{
  m_nAnimState.kill();
  m_nY = -SIZE_METEOR_SPRITE;
  PA_DualStopSpriteAnim(m_nMeteor);
  PA_DualSetSpriteAnimFrame(m_nMeteor, 0);
  PA_DualSetSpriteXY(m_nMeteor, getMeteorX(), getMeteorY());
  PA_DualSetSpriteXY(m_nContent, getContentX(), getContentY());
}

//--------------------------------------------------------------------

void
Meteor::pause()
{
  m_bPaused = true;

  // move off screen.
  PA_DualSetSpriteXY(m_nMeteor, getMeteorX(), 384);
  PA_DualSetSpriteXY(m_nContent, getContentX(), 384);

  // pause any animations.
  PA_DualSpriteAnimPause(m_nMeteor, 1);
  m_nFrame = PA_DualGetSpriteAnimFrame(m_nMeteor);
}

//--------------------------------------------------------------------

void
Meteor::unpause()
{
  m_bPaused = false;

  // move back onto screen.
  PA_DualSetSpriteXY(m_nMeteor, getMeteorX(), getMeteorY());
  PA_DualSetSpriteXY(m_nContent, getContentX(), getContentY());

  // unpause any animations.
  m_nAnimState.restart();

  // reset last tick.
  m_nLastTick = Tick(m_nTimerID);
}

//--------------------------------------------------------------------

void
Meteor::addSignals(SignalSet &signals)
{
  u16 binary(0);

  if (m_bPaused) { return; }

  if (HitBox::intersect(getHitBox(), getBottomHitBox()).valid()
      && !exploding())
  {
    startExplodeAnimation();
    binary = binary | (1 << MeteorBottomCollision);
    signals.insert(MeteorBottomCollision);
  }
  if (exploding())
  {
    binary = binary | (1 << MeteorsExploding);
    signals.insert(MeteorsExploding);
  }
  if (refreshing())
  {
    binary = binary | (1 << MeteorsRefreshing);
    signals.insert(MeteorsRefreshing);
  }

  if (atComboBegin() && m_bSendComboSignals)
  {
    signals.insert(ShowCombo);
  }

  if (atExplodeAnimationEnd() && m_bSendComboSignals)
  {
    signals.insert(HideCombo);
    m_bSendComboSignals = false;
  }
}

//--------------------------------------------------------------------

void
Meteor::handleSignals(const SignalSet &signals)
{
  if (signals.contains(Unpause))
  {
    unpause();
    return;
  }

  if (m_bPaused) { return; }

  if (signals.contains(Pause))
  {
    pause();
    return;
  }

  if (signals.contains(RefreshNeeded)
      && !signals.contains(MeteorsExploding)
      && !signals.contains(MissileExploding)
      && !exploding()
      && m_nCollisionType == None)
  {
    startRefreshAnimation();
  }

  if (signals.contains(MissileExplosionDone))
  {
    if (m_nCollisionType == Correct) { startExplodeAnimation(); }
    else if (m_nCollisionType == Incorrect 
	     || m_nCollisionType == RealAnswer)
    {
      startWrongAnimation();
    }
  }

  if (signals.contains(GameOver))
  {
    reset();
    m_nState = Paused;
  }
}

//--------------------------------------------------------------------

void
Missile::hideKana()
{
  for (u8 i=0; i < NUM_READING_SPRITES; ++i)
  {
    PA_DualSetSpriteXY(SPRITE_BEGIN_KANA + i,
		       16 * i, 384);
  }
}

//--------------------------------------------------------------------

void
Missile::emptyBar()
{
  for (u8 i = 0; i < NUM_MISSILE_BAR_SPRITES; ++i)
  {
    PA_DualSetSpriteAnimFrame(SPRITE_BEGIN_MISSILE_BAR + i, 0);
    PA_DualStopSpriteAnim(SPRITE_BEGIN_MISSILE_BAR + i);
  }
}

//--------------------------------------------------------------------

void
Missile::resetLocation()
{
  m_nY = 350;
  PA_DualSetSpriteXY(SPRITE_MISSILE, m_nX, m_nY);
  PA_DualSetSpriteXY(SPRITE_EXPLOSION, 0, 384);
  m_nState = Waiting;
}

//--------------------------------------------------------------------

void
Missile::pause()
{
  PA_DualSetSpriteXY(SPRITE_MISSILE, m_nX, 384);
  PA_DualSpriteAnimPause(SPRITE_MISSILE, 1);
  if (m_nState == GoodCollision)
  {
    PA_DualSetSpriteXY(SPRITE_EXPLOSION, m_nX, 384);
  }
  m_bPaused = true;
}

//--------------------------------------------------------------------

void
Missile::unpause()
{
  PA_DualSetSpriteXY(SPRITE_MISSILE, m_nX, m_nY);
  PA_DualSpriteAnimPause(SPRITE_MISSILE, 0);
  if (m_nState == GoodCollision)
  {
    PA_DualSetSpriteXY(SPRITE_EXPLOSION, m_nX, m_nY);
  }
  m_bPaused = false;
}

//--------------------------------------------------------------------

void
Missile::move()
{
  if (m_bPaused) { return; }

  u8 frame(0);
  switch (m_nState)
  {
  case InAir:
    m_nY -= 6;
    if (m_nY < -HEIGHT_MISSILE_SPRITE)
    { 
      m_nY = 350;
      m_nState = Ready;
    }
    PA_DualSetSpriteXY(SPRITE_MISSILE, m_nX, m_nY);
    break;
  case GoodCollision:
    frame = PA_DualGetSpriteAnimFrame(SPRITE_MISSILE);
    m_nY = m_nTargetY + HEIGHT_MISSILE_SPRITE - (frame * EXPLODE_FRAME_DY);
    PA_DualSetSpriteXY(SPRITE_MISSILE, m_nX, m_nY);
    break;
  case BadCollision:
    break;
  case Load:
    if (m_bNeedLoad) { startLoad(); }
    if ((PA_DualGetSpriteAnimFrame(SPRITE_BEGIN_MISSILE_BAR
				   + m_nBarRefreshing)
	 == 4)
	&& m_nBarRefreshing < 3)
    {
      PA_DualStopSpriteAnim(SPRITE_BEGIN_MISSILE_BAR + m_nBarRefreshing);
      PA_DualSetSpriteAnim(SPRITE_BEGIN_MISSILE_BAR + m_nBarRefreshing, 4);

      m_nBarRefreshing += 1;
      PA_DualStartSpriteAnimEx(SPRITE_BEGIN_MISSILE_BAR + m_nBarRefreshing,
			       0, 4, 30, ANIM_UPDOWN, 1);
    }
    if (atLoadAnimationEnd())
    { 
      showReading();
      m_nState = Ready;
      unlock();
    }
    break;
  case Reload:
    if (m_reading.size())
    {
      for (u8 i=0; i < NUM_MISSILE_BAR_SPRITES; ++i)
      {
	PA_DualStopSpriteAnim(SPRITE_BEGIN_MISSILE_BAR + i);
	PA_DualSetSpriteAnim(SPRITE_BEGIN_MISSILE_BAR + i, 4);
      }
      showReading();
      m_nState = Ready;
    }
    break;
  case Ready:
    if (m_bNeedLoad) { startLoad(); }
    else if (m_bNeedReload) { startReload(); }
    break;
  case Waiting:
    if (m_bNeedLoad) { startLoad(); }
    else if (m_bNeedReload) { startReload(); }
    break;
  default:
    break;
  }
}

//--------------------------------------------------------------------

void
Missile::fire(s32 x)
{
  if (m_nState == Ready)
  {
    m_nX = x;
    m_nState = InAir;
    emptyBar();
    hideKana();
  }
}

//--------------------------------------------------------------------

void
Missile::startGoodCollision()
{
  m_nTargetY = m_nY - HEIGHT_MISSILE_SPRITE;
  m_nState = GoodCollision;
  PA_DualStartSpriteAnimEx(SPRITE_MISSILE, 0, 4, 10, ANIM_ONESHOT);

  PA_DualSetSpriteXY(SPRITE_EXPLOSION, m_nX, m_nY);
  PA_DualStartSpriteAnimEx(SPRITE_EXPLOSION, 0, 3, 10, ANIM_LOOP, 2);
}


//--------------------------------------------------------------------

void
Missile::startBadCollision()
{
  m_nState = BadCollision;
  hideKana();
  emptyBar();
  PA_DualStartSpriteAnimEx(SPRITE_MISSILE, 0, 4, 10, ANIM_ONESHOT);
}

//--------------------------------------------------------------------

void
Missile::startLoad()
{
  if (m_nState != Load)
  {
    m_nState = Load;
    m_bNeedLoad=false;
    m_nBarRefreshing=0;
    for (u8 i=0; i < NUM_MISSILE_BAR_SPRITES; ++i)
    {
      PA_DualSetSpriteAnimFrame(SPRITE_BEGIN_MISSILE_BAR + i, 0);
    }
    PA_DualStartSpriteAnimEx(SPRITE_BEGIN_MISSILE_BAR, 0, 4, 30,
			     ANIM_UPDOWN, 1);
    m_nLoadCount += 1;
  }
}

//--------------------------------------------------------------------

void
Missile::startReload()
{
  if (m_nState != Reload)
  {
    m_nState = Reload;
    m_bNeedReload = false;
    for (u8 i=0; i < NUM_MISSILE_BAR_SPRITES; ++i)
    {
      PA_DualSetSpriteAnimFrame(SPRITE_BEGIN_MISSILE_BAR + i, 0);
      PA_DualStartSpriteAnimEx(SPRITE_BEGIN_MISSILE_BAR + i,
			       5, 8, 30, ANIM_UPDOWN, 1);
    }
  }
}

//--------------------------------------------------------------------

void
Missile::addSignals(SignalSet &signals)
{
  u16 binary(0);

  if (m_bPaused) { return; }

  // Somehow atAnimationEnd() returns true in other states, too.
  if (atAnimationEnd()
      && (m_nState == GoodCollision || m_nState == BadCollision))
  {
    binary = binary | (1 << MissileExplosionDone);
    signals.insert(MissileExplosionDone);
    resetLocation();
    lock();
  }

  if (m_nState == Load || m_nState == Reload)
  {
    signals.insert(MissileReloading);
  }

  if (exploding())
  {
    binary = binary | (1 << MissileExploding);
    signals.insert(MissileExploding);
  }
}

//--------------------------------------------------------------------

void
Missile::handleSignals(const SignalSet &signals)
{
  if (signals.contains(Unpause))
  {
    unpause();
    return;
  }

  if (signals.contains(Pause))
  {
    pause();
  }

  if (m_bPaused) { return; }

  if (!signals.contains(MeteorsRefreshing)
      && m_setPreviousSignals.contains(MeteorsRefreshing))
  {
    setLoad();
    unlock();
  }

  if (signals.contains(MissileGoodCollision))
  {
    startGoodCollision();
  }
  
  if (signals.contains(MissileBadCollision))
  {
    startBadCollision();
  }

  m_setPreviousSignals = signals;
}

//--------------------------------------------------------------------

void
Missile::setReading(Reading s)
{
  m_reading = Reading(s);
  for (u8 i=0; i < NUM_READING_SPRITES; ++i)
  {
    PA_DualSetSpriteXY(SPRITE_BEGIN_KANA + i, (16 * i), 384);
  }
}

//--------------------------------------------------------------------

void
Missile::showReading()
{
  for (u8 i=0; i < m_reading.size(); ++i)
  {
    PA_DualUpdateSpriteGfx(SPRITE_BEGIN_KANA + i,
			   (void *)getKana(m_reading[i]));
    PA_DualSetSpriteXY(SPRITE_BEGIN_KANA + i,
		       (16 * i), 368);
  }
}

//--------------------------------------------------------------------

ComboGraphic::ComboGraphic(u8 background, u8 begin_digit)
  : m_nX(0), m_nY(0),
    m_nBackgroundSprite(background),
    m_nBeginDigitSprites(begin_digit)
{
  hide();
}

//--------------------------------------------------------------------

void
ComboGraphic::setSpriteLocation(s32 x, s32 y)
{
  PA_DualSetSpriteXY(m_nBackgroundSprite, x, y);
  for (u8 i=0; i < NUM_DIGIT_SPRITES; ++i)
  {
    PA_DualSetSpriteXY(m_nBeginDigitSprites+i,
		       x + 32 - m_nRightOffset - (8 * (i+1)), y + 12);
  }
}

//--------------------------------------------------------------------

void ComboGraphic::show()
{
  setSpriteLocation(m_nX, m_nY);
}

//--------------------------------------------------------------------

void ComboGraphic::hide()
{
  setSpriteLocation(m_nX, 384);
}

//--------------------------------------------------------------------

void
ComboGraphic::setLocation(s32 x, s32 y)
{
  m_nX = x;
  m_nY = y;
}

//--------------------------------------------------------------------

void
ComboGraphic::setCount(u16 c)
{
  if (c < 10) { m_nRightOffset = 12; }
  else if (c < 100) { m_nRightOffset = 8; }
  else { m_nRightOffset = 4; }

  bool non_zero(false);
  for (u8 i=0; i < NUM_DIGIT_SPRITES; ++i)
  {
    u8 j = NUM_DIGIT_SPRITES - i - 1;
    u16 power_of_ten=1;
    for (u8 k=0; k < j; ++k) { power_of_ten = power_of_ten * 10; }
    u8 digit = c / power_of_ten;
    c -= (digit * power_of_ten);
    if (digit) { non_zero=true; }

    if (non_zero)
    {
      PA_DualSetSpriteAnim(m_nBeginDigitSprites+j, digit);
    }
    else
    {
      PA_DualSetSpriteAnim(m_nBeginDigitSprites+j, 10);
    }
  }

  if (c < 10)       { PA_DualSetSpriteAnim(m_nBackgroundSprite, 0); }
  else if (c < 20)  { PA_DualSetSpriteAnim(m_nBackgroundSprite, 1); }
  else if (c < 30)  { PA_DualSetSpriteAnim(m_nBackgroundSprite, 2); }
  else if (c < 50)  { PA_DualSetSpriteAnim(m_nBackgroundSprite, 3); }
  else if (c < 100) { PA_DualSetSpriteAnim(m_nBackgroundSprite, 4); }
  else              { PA_DualSetSpriteAnim(m_nBackgroundSprite, 5); }
}

//--------------------------------------------------------------------

void
ComboGraphic::initSprites(u8 background, u8 begin_digit)
{
  PA_DualCreateSprite(background,(void*)combo_Sprite,
		      OBJ_SIZE_32X32,1, PALETTE_GAMEPLAY,
		      0, 384);
  PA_DualSetSpritePrio(background, 1);

  for (u8 i=0; i < NUM_DIGIT_SPRITES; ++i)
  {
    PA_DualCreateSprite(begin_digit+i,(void*)digits_Sprite,
			OBJ_SIZE_8X8,1, PALETTE_GAMEPLAY,
			0, 384);
    PA_DualSetSpritePrio(begin_digit+i, 0);
  }
}

//--------------------------------------------------------------------

GameLoop::GameLoop()
  : m_combo(SPRITE_COMBO_BACKGROUND, SPRITE_BEGIN_COMBO_DIGIT),
    m_nLifePoints(16), m_nCursorX(0),
    m_bRefreshData(false), m_nPreviousMeteor(0),
    m_bPaused(false)
{
  StartTime(true);
  m_nTimerID = NewTimer(true);

  m_menuPause.setStyle(MenuInterface::Style(MenuInterface::Style::Centered,
					    0, 24, 2, 4));
  m_menuPause.addTextItem(ToGame, "Continue game");
  m_menuPause.addTextItem(ToMenu, "Return to menu");

  cleanup();
}

//--------------------------------------------------------------------

void
GameLoop::initSprites()
{
  PA_DualLoadSpritePal(1, (void*)gameplay_Pal);  

  PA_DualLoadTiledBg(1, bg2);
  PA_DualLoadTiledBg(3, bg3);

  PA_DualCreateSprite(SPRITE_CURSOR,(void*)cursor_Sprite,
		      OBJ_SIZE_32X16,1, PALETTE_GAMEPLAY,
		      0, 344);
  PA_DualSetSpritePrio(SPRITE_CURSOR, 0);

  PA_DualCreateSprite(SPRITE_MISSILE,(void*)missile_Sprite,
		      OBJ_SIZE_32X64,1, PALETTE_GAMEPLAY,
		      0, 350);
  PA_DualSetSpritePrio(SPRITE_MISSILE, 2);

  PA_DualCreateSprite(SPRITE_EXPLOSION,(void*)explosion_Sprite,
		      OBJ_SIZE_32X16,1, PALETTE_GAMEPLAY,
		      0, -32);
  PA_DualSetSpritePrio(SPRITE_EXPLOSION, 0);

  for (u8 i = SPRITE_BEGIN_MISSILE_BAR; i < SPRITE_BEGIN_KANA; ++i)
  {
    PA_DualCreateSprite(i,(void*)missile_bar_Sprite,
			OBJ_SIZE_32X16,1, PALETTE_GAMEPLAY,
			32 * (i - SPRITE_BEGIN_MISSILE_BAR), 368);
    PA_DualSetSpritePrio(i, 1);
  }

  for (u8 i=0; i < NUM_LIFE_BAR_SPRITES; ++i)
  {
    PA_DualCreateSprite(SPRITE_BEGIN_LIFE_BAR + i,
			(void*)life_bar_Sprite,
			OBJ_SIZE_16X16,1, PALETTE_GAMEPLAY,
			256 - (16 * (i + 1)), 368);
    PA_DualSetSpritePrio(SPRITE_BEGIN_LIFE_BAR + i, 0);
    PA_DualSetSpriteAnim(SPRITE_BEGIN_LIFE_BAR+i, 0);
  }

  PA_DualCreateSprite(SPRITE_MISSILE_ICON,
		      (void*)missile_icon_Sprite,
		      OBJ_SIZE_16X16,1, PALETTE_GAMEPLAY,
		      128, 368);
  PA_DualSetSpritePrio(SPRITE_MISSILE_ICON, 0);

  PA_DualCreateSprite(SPRITE_LIFE_ICON,
		      (void*)life_icon_Sprite,
		      OBJ_SIZE_16X16,1, PALETTE_GAMEPLAY,
		      176, 368);
  PA_DualSetSpritePrio(SPRITE_LIFE_ICON, 0);

  /*
  for (u8 i=SPRITE_BEGIN_KANA; i < SPRITE_BEGIN_METEOR; ++i)
  {
    PA_DualCreateSprite(i,(void*)u3050_Sprite,
			OBJ_SIZE_16X16,0, PALETTE_CHARS,
			16 * (i - SPRITE_BEGIN_KANA), 384);
    PA_DualSetSpritePrio(i, 0);
  }
  */

  for (u8 i=0; i < NUM_METEORS; ++i)
  {
    /*
    PA_DualCreateSprite(SPRITE_BEGIN_METEOR_CONTENT + i,
			NULL,
			OBJ_SIZE_32X32,0, PALETTE_CHARS,
			0, 384);
    PA_DualSetSpritePrio(SPRITE_BEGIN_METEOR_CONTENT + i, 2);
    */

    PA_DualCreateSprite(SPRITE_BEGIN_METEOR + i,
			(void*)meteor_back_Sprite,
			OBJ_SIZE_64X64,1, PALETTE_GAMEPLAY, 
			0, 384);
    PA_DualSetSpritePrio(SPRITE_BEGIN_METEOR + i, 3);

  }

  ComboGraphic::initSprites(SPRITE_COMBO_BACKGROUND,
			    SPRITE_BEGIN_COMBO_DIGIT);
}

//--------------------------------------------------------------------

void
GameLoop::init(u8 num_meteors, u32 meteor_tpm)
{
  m_vecMeteors.clear();
  if (!num_meteors) { return; }

  std::vector<u32> offsets;
  for (u8 i=0; i < num_meteors; ++i) { offsets.push_back(32 * i); }
  random_shuffle(offsets.begin(), offsets.end());

  u16 total_space = 256 - (num_meteors * 32);
  u16 space = (total_space / num_meteors);
  for (u8 i =0; i < num_meteors; ++i)
  {
    Meteor m(((32 + space) * i) + ((space/2) - 16), -48,
	     SPRITE_BEGIN_KANJI + i,
	     SPRITE_BEGIN_METEOR + i,
	     meteor_tpm, m_nTimerID,
	     offsets[i], 1, 3 * i);
    m_vecMeteors.push_back(m);
  }

  PA_DualSetSpriteXY(SPRITE_MISSILE_ICON, 128, 368);  
  PA_DualSetSpriteXY(SPRITE_LIFE_ICON, 176, 368);  

  for (u8 i=0; i < NUM_LIFE_BAR_SPRITES; ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_LIFE_BAR + i, 368);
  }

  for (u8 i = 0; i < NUM_MISSILE_BAR_SPRITES; ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_MISSILE_BAR + i, 368);
  }

}

//-------------------------------------------------------------------

void
GameLoop::cleanup()
{
  // maybe requires fixing.
  for (u8 i=SPRITE_CURSOR; i < SPRITE_BEGIN_METEOR+12; ++i)
  {
    PA_DualSetSpriteY(i, 384);
  }

  for (u8 i=0; i < NUM_READING_SPRITES; ++i)
  {
    PA_DualSetSpriteY(SPRITE_BEGIN_KANA + i, 384);
  }

  m_menuPause.hide();

  PA_DualHideBg(1);
  PA_DualHideBg(3);
}

//-------------------------------------------------------------------


u8
GameLoop::indexMissileCollision()
{
  u8 res = m_vecMeteors.size();
  for (u8 i = 0; i < m_vecMeteors.size(); ++i)
  {
    if (HitBox::intersect(m_vecMeteors[i].getHitBox(),
			  m_missile.getHitBox()).valid()
	&& !(m_vecMeteors[i].exploding()))
    {
      res = i;
    }
  }
  return res;
}

//-------------------------------------------------------------------

u16
GameLoop::missileRowMatch(u8 meteor_ind)
{
  u16 res(ContentData::BAD_ROW);
  u16 row_ind = m_vecMeteors[meteor_ind].getContentRowIndex();
  std::pair<u16, u16> range(GameData::content.getCharRange(row_ind));
  for (u16 i=range.first; i < range.second; ++i)
  {
    if (GameData::content.getRow(i).reading() == m_missile.reading())
    {
      res = i;
    }
  }
  return res;
}

//-------------------------------------------------------------------

// has the missile hit a meteor?
void
GameLoop::checkMissileCollision(SignalSet &signals)
{
  static u16 num_collisions(0);

  if (!m_missile.inAir()) { return; }

  u8 index = indexMissileCollision();
  if (index < m_vecMeteors.size())
  {
    num_collisions += 1;
    
    // check if missile reading
    // is among readings for this meteor.
    u16 row_ind = missileRowMatch(index);
    if (row_ind != ContentData::BAD_ROW)
    {
      m_nCombo += 1;

      // relocate combo graphics using _index_
      m_combo.setLocation(m_vecMeteors[index].getContentX(),
			  m_vecMeteors[index].getContentY());
      m_combo.setCount(m_nCombo);

      // different animation?
      GameData::record.lowerMasteredCutoff();
      RecordEditRes er(GameData::record.recordCorrect(row_ind));
      // if mastered, mark for different animation?
      signals.insert(MissileGoodCollision);
      m_vecMeteors[index].setCollisionType(Meteor::Correct);
      m_vecMeteors[index].activateComboSignals();
      if (m_nLifePoints < 16)
      {
	m_nLifePoints += 1;
	drawLifeBar();
      }
    }
    else
    {
      m_nCombo=0;

      std::vector<u16> wrong_rows;
      wrong_rows.push_back(m_vecMeteors[index].getContentRowIndex());

      m_vecMeteors[index].setCollisionType(Meteor::Incorrect);

      for (u8 i=0; i < m_vecMeteors.size(); ++i)
      {
	u16 row_ind2 = missileRowMatch(i);
	if (i != index && row_ind2 != ContentData::BAD_ROW)
	{
	  m_vecMeteors[i].setCollisionType(Meteor::RealAnswer);
	  wrong_rows.push_back(row_ind2);
	}
      }

      for (u16 j=0; j < wrong_rows.size(); ++j)
      {
	GameData::record.recordCorrect(wrong_rows[j]);
      }

      signals.insert(MissileBadCollision);
    }
    signals.insert(MissileExploding);
    m_bRefreshData = true;
  }
}

//-------------------------------------------------------------------

void
GameLoop::moveMeteors()
{
  for (MeteorVec::iterator iter = m_vecMeteors.begin();
       iter != m_vecMeteors.end(); ++iter)
  {
    iter->move();
  }
}

//-------------------------------------------------------------------

void
GameLoop::moveCursor()
{
  if (m_bPaused) { return; }

  if (Pad.Held.Right - Pad.Held.Left > 0) { m_nCursorX += 4; }
  else if (Pad.Held.Left - Pad.Held.Right > 0) { m_nCursorX -= 4; }
  else if (Stylus.Held && (Stylus.Y < 172) && (Stylus.Y > 152))
  {
    s32 cursor_centerx = m_nCursorX + (WIDTH_CURSOR_SPRITE / 2);
    s32 dx = Stylus.X - cursor_centerx;
    if (abs(dx) < DIST_STYLUS_CURSOR)
    {
      if (dx > 0) { m_nCursorX += 6; }
      else if (dx < 0) { m_nCursorX -= 6; }
    }
  }

  if (m_nCursorX < 0) { m_nCursorX = 0; }
  if (m_nCursorX > 255 - WIDTH_CURSOR_SPRITE)
  {
    m_nCursorX = 255 - WIDTH_CURSOR_SPRITE;
  }

  PA_DualSetSpriteXY(SPRITE_CURSOR, m_nCursorX, 350);
}

//-------------------------------------------------------------------

void
GameLoop::drawLifeBar()
{
  for (u8 i=0; i<NUM_LIFE_BAR_SPRITES; ++i)
  {
    s8 x = m_nLifePoints - (4 * i);
    if (x >= 4)
    {
      PA_DualSetSpriteAnim(SPRITE_BEGIN_LIFE_BAR+i, 4);
    }
    else if (x <= 0)
    {
      PA_DualSetSpriteAnim(SPRITE_BEGIN_LIFE_BAR+i, 0);
    }
    else
    {
      PA_DualSetSpriteAnim(SPRITE_BEGIN_LIFE_BAR+i, x);
    }
  }
}

//-------------------------------------------------------------------

void
GameLoop::handleMissileInput()
{
  if (m_bPaused) { return; }

  if (m_missile.ready() && m_missile.unlocked())
  {
    if (Pad.Newpress.A
	|| (Stylus.Newpress
	    && Stylus.Y < 132
	    && (abs(Stylus.X
		    - (m_nCursorX
		       + (WIDTH_CURSOR_SPRITE/2)))
		< DIST_STYLUS_FIRE)))
    {
      m_missile.fire(m_nCursorX);
    }
    else if (Pad.Newpress.B
	     || (Stylus.Newpress && Stylus.Y < 132))
    {
      resetMissile();
      m_missile.setReload();
    }
  }
}

//-------------------------------------------------------------------

void
GameLoop::resetMeteors()
{
  std::vector<u16> excluded;
  u8 print_row=0;
  for (MeteorVec::iterator iter = m_vecMeteors.begin();
       iter != m_vecMeteors.end();
       ++iter)
  {
    u16 row = GameData::record.pick(excluded);
    print_row += 1;
    iter->setContentRow(row);
    std::pair<u16, u16> range(GameData::content.getCharRange(row));
    for (u16 i=range.first; i < range.second; ++i)
    {
      excluded.push_back(i);
    }
  }
}

//-------------------------------------------------------------------

void
GameLoop::resetMissile()
{
  u8 meteor_ind = m_vecMeteors.size();
  if (m_vecChoosableMeteors.empty())
  {
    // fill vec with every index.
    // randomize.
    // take last one.

    for (u8 i=0; i < m_vecMeteors.size(); ++i)
    {
      m_vecChoosableMeteors.push_back(i);
    }
    random_shuffle(m_vecChoosableMeteors.begin(),
		   m_vecChoosableMeteors.end());


    meteor_ind = m_vecChoosableMeteors.back();
    m_vecChoosableMeteors.pop_back();
  }
  else if (m_vecChoosableMeteors.size() == 1)
  {
    // new meteor ind is the only one left.
    // fill vec with every index but that one.
    // randommize vec.

    meteor_ind = m_vecChoosableMeteors[0];
    m_vecChoosableMeteors.clear();

    for (u8 i=0; i < m_vecMeteors.size(); ++i)
    {
      m_vecChoosableMeteors.push_back(i);
    }
    random_shuffle(m_vecChoosableMeteors.begin(),
		   m_vecChoosableMeteors.end());

  }
  else
  {
    // take last one.
    meteor_ind = m_vecChoosableMeteors.back();
    m_vecChoosableMeteors.pop_back();
  }

  u16 row_ind = m_vecMeteors[meteor_ind].getContentRowIndex();
  m_missile.setReading(GameData::content.getRow(row_ind).reading());
}

//-------------------------------------------------------------------

void
GameLoop::addSignals(SignalSet &signals)
{
  if (Pad.Newpress.Start && !m_bPaused)
  {
    m_menuPause.show();
    signals.insert(Pause);
    m_bPaused = true;
  }

  if (m_bPaused)
  {
    m_menuPause.handleStandardInput();
    if (m_menuPause.clickEvent())
    {
      if (m_menuPause.currentIndex() == ToGame)
      {
	m_bPaused = false;
	m_menuPause.hide();
	signals.insert(Unpause);
      }
      else if (m_menuPause.currentIndex() == ToMenu)
      {
	signals.insert(GameOver);
      }
    }

    return;
  }

  checkMissileCollision(signals);
  if (m_bRefreshData) { signals.insert(RefreshNeeded); }
  if (m_nLifePoints <= 0) { signals.insert(GameOver); }
}

//-------------------------------------------------------------------

void
GameLoop::handleSignals(const SignalSet &signals)
{
  if (m_bPaused) { return; }

  if (signals.contains(Unpause)) { return; }

  if (signals.contains(MeteorBottomCollision))
  {
    m_nCombo=0;
    m_bRefreshData = true;
    m_nLifePoints -= 8;
  }

  if (!signals.contains(MeteorsRefreshing)
      //      && !signals.contains(MeteorsExploding)
      && m_setPreviousSignals.contains(MeteorsRefreshing))
  {
    resetMeteors();
    resetMissile();
  }

  if (signals.contains(RefreshNeeded)
      && !signals.contains(MeteorsExploding)
      && !signals.contains(MissileExploding))
  {
    m_missile.lock();
    m_bRefreshData = false;
  }

  if (signals.contains(ShowCombo) && m_nCombo >5)
  {
    m_combo.show();
  }

  if (signals.contains(HideCombo))
  {
    m_combo.hide();
  }

  m_setPreviousSignals = signals;
}

//-------------------------------------------------------------------

bool
GameLoop::gameplayCycle()
{
  static u32 previous_binary(0);

  if (PA_CheckLid())
  {
    m_bDSClosed = true;
    return true;
  }

  if (!PA_CheckLid() && m_bDSClosed)
  {
    // reset meteors' last tick.
    for (MeteorVec::iterator iter = m_vecMeteors.begin();
	 iter != m_vecMeteors.end();
	 ++iter)
    {
      iter->resetLastTick();
    }
    m_bDSClosed = false;
    return true;
  }

  drawLifeBar();

  SignalSet signals;
  addSignals(signals);
  m_missile.addSignals(signals);
  for (MeteorVec::iterator iter = m_vecMeteors.begin();
       iter != m_vecMeteors.end();
       ++iter)
  {
    iter->addSignals(signals);
  }

  if (DEBUG)
  {
    u32 binary = signals.binary();
    PA_OutputText(1, 0, 22, "                        ");
    for (u8 i=0; i < 20; ++i)
    {
      PA_OutputText(1, i, 23, "%d", (u8)signals.contains(i));
    }
    if (binary != previous_binary)
    {
      m_vecSignals.push_back(binary);
      previous_binary = binary;
    }
  }

  handleSignals(signals);
  m_missile.handleSignals(signals);
  for (MeteorVec::iterator iter = m_vecMeteors.begin();
       iter != m_vecMeteors.end();
       ++iter)
  {
    iter->handleSignals(signals);
  }

  moveMeteors();
  m_missile.move();
  if (m_nLifePoints > 0)
  {
    moveCursor();
    handleMissileInput();
  }

  return !signals.contains(GameOver);
}

//------------------------------------------------------------------

bool
GameLoop::pauseCycle()
{
  return false;
}

//------------------------------------------------------------------

void
GameLoop::run()
{
  bool loop(true);

  PA_DualShowBg(1);
  PA_DualShowBg(3);

  m_missile = Missile();
  resetMeteors();
  resetMissile();
  m_nLifePoints = 16;
  m_bRefreshData = false;
  m_nCombo=0;
  m_bPaused = false;
  m_missile.setLoad();
  drawLifeBar();

  while (loop)
  {
    loop = gameplayCycle();
    PA_WaitForVBL();
  }
  cleanup();

  if (DEBUG)
  {
    // write signals to file.
    DataFile file;
    file.addVec("signals", m_vecSignals);
    file.write("gameplay_log.txt");
  }
}
