//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com

#ifndef GAMEPLAY_H
#define GAMEPLAY_H

#include <PA9.h>
#include <set>
#include <vector>
#include "consts.h"
#include "data.h"
#include "menu_base.h"

class HitBox
{
 private:
  s32 x0;
  s32 y0;
  s32 x1;
  s32 y1;

 public:
 HitBox(s32 a=0, s32 b=0, s32 c=0, s32 d=0)
   : x0(a), y0(b), x1(c), y1(d)
  {}

  static HitBox intersect(HitBox const& a, HitBox const& b)
  {
    return HitBox((a.x0 < b.x0 ? b.x0 : a.x0),
		  (a.y0 < b.y0 ? b.y0 : a.y0),
		  (a.x1 < b.x1 ? a.x1 : b.x1),
		  (a.y1 < b.y1 ? a.y1 : b.y1));
  }

  bool valid() { return (x1 > x0) && (y1 > y0); }

  void display(u8 screen, u8 y)
  {
    PA_OutputText(screen,0,y,"(%d, %d), (%d, %d)",
		  x0, y0, x1, y1);

  }
};

enum GameplaySignal
{
  MeteorBottomCollision,
  MissileGoodCollision,
  MissileBadCollision,
  MissileCollisionDone,  // obsolete
  MeteorsExploding,
  MeteorsRefreshing,
  MeteorRefreshDone,  // obsolete
  MissileExploding,
  MissileExplosionDone,
  MissileReloading,
  RefreshNeeded,
  ShowCombo,
  HideCombo,
  Pause,
  Unpause,
  GameOver
};

class SignalSet
{
 private:
  std::set<GameplaySignal> m_setData;

 public:
  bool contains(GameplaySignal gs) const
  {
    return (m_setData.find(gs) != m_setData.end());
  }
  bool contains(u8 i) const { return contains((GameplaySignal)i); }
  void insert(GameplaySignal gs) { m_setData.insert(gs); }
  u32 binary() const
  {
    u32 res(0);
    for (u8 i=0; i < 32; ++i) { if (contains(i)) { res = res | (1 << i); } }
    return res;
  }
};


class AnimationState
{
 private:
  u8 m_nSprite;
  u8 m_nMinFrame;
  u8 m_nMaxFrame;
  u8 m_nCycles;
  u8 m_nSpeed;
  bool m_bAlive;

  u8 m_nCurrentFrame;
  u8 m_nCurrentCycle;

 public:
  AnimationState(u8 s, u8 minf, u8 maxf, u8 c, u8 speed);
  void reset();
  void update();
  void start();
  void restart();
  void kill();
  void display(u8 screen, u8 line);
  bool alive() const { return m_bAlive; }
};

class Meteor
{
 private:
  static const u8 SIZE_METEOR_SPRITE = 64;
  static const u8 SIZE_CONTENT_SPRITE = 32;

  enum State
  {
    Normal,
    Refresh,
    Exploding,
    StateBadAnswer,
    StateRealAnswer,
    Paused
  };

 public:
  enum CollisionType
  {
    None,
    Correct,
    Incorrect,
    RealAnswer
  };

 private:
  s32 m_nX;
  s32 m_nY;
  u8 m_nContent;
  u8 m_nMeteor;
  bool m_bExploding;
  u32 m_nTicksPerMove;
  u8 m_nTimerID;
  u32 m_nLastTick;
  State m_nState;
  u32 m_nRefreshStartTime;
  u16 m_nRowIndex;
  CollisionType m_nCollisionType;
  u16 m_nInitialOffset;
  bool m_bSendComboSignals;
  bool m_bPaused;
  u8 m_nFrame;
  AnimationState m_nAnimState;
  u8 m_nOutputScreen;
  u8 m_nOutputLine;

 private:
  HitBox getBottomHitBox() { return HitBox(0, 340, 255, 500); }

 private:

  void startExplodeAnimation();
  void startRefreshAnimation();
  void startWrongAnimation();
  bool atExplodeAnimationEnd() const
  {
    return (PA_DualGetSpriteAnimFrame(m_nMeteor) > 4);
  }

  bool atRefreshAnimationEnd() const
  {
    return (Tick(m_nTimerID) - m_nRefreshStartTime > 500);
  }

  bool atComboBegin() const
  {
    return (PA_DualGetSpriteAnimFrame(m_nMeteor) == 2);
  }

  void reset();
  void pause();
  void unpause();

 public:
  Meteor(s32 a, s32 b, u8 sc, u8 sm, u32 tpm, u8 tid, u16 offset,
	 u8 screen, u8 line);
  void init();
  void move();

  HitBox getHitBox() const
  {
    s32 half_width = SIZE_CONTENT_SPRITE / 2;
    return HitBox(getCenterX() - half_width,
		  getCenterY() - half_width,
		  getCenterX() + half_width,
		  getCenterY() + half_width);
  }
  bool exploding() const { return (m_nState == Exploding); }
  bool refreshing() const { return (m_nState == Refresh); }
  u16 getContentRowIndex() const { return m_nRowIndex; }

  void setCollisionType(CollisionType ct) { m_nCollisionType = ct; }
  void setContentRow(u16 row_ind);
  void activateComboSignals() { m_bSendComboSignals=true; }
  void addSignals(SignalSet &signals);
  void handleSignals(const SignalSet &signals);

  // Geometry functions.
  s32 getMeteorX() const { return m_nX; }
  s32 getMeteorY() const { return m_nY; }
  s32 getContentX() const
  {
    return getCenterX() - (SIZE_CONTENT_SPRITE / 2);
  }
  s32 getContentY() const
  {
    return getCenterY() - (SIZE_CONTENT_SPRITE / 2);
  }

  // The center functions assume the meteor sprite
  // is at least as big as the content sprite,
  // and not the other way around.
  s32 getCenterX() const { return m_nX + (SIZE_METEOR_SPRITE / 2); }
  s32 getCenterY() const { return m_nY + (SIZE_METEOR_SPRITE / 2); }

  void resetLastTick();

};

class Missile
{
 private:
  static const u8 HEIGHT_MISSILE_SPRITE=64;
  static const u8 WIDTH_MISSILE_SPRITE=32;
  static const s32 EXPLODE_FRAME_DY=16;

 private:
  enum State
  {
    Ready,
    Waiting,
    Load,
    Reload,
    InAir,
    GoodCollision,
    BadCollision
  };

 private:
  s32 m_nX;
  s32 m_nY;
  State m_nState;
  bool m_bLocked;
  s32 m_nTargetY;
  u8 m_nLoadCount;
  u8 m_nBarRefreshing;
  bool m_bNeedLoad;
  bool m_bNeedReload;
  Reading m_reading;
  bool m_bPaused;
  SignalSet m_setPreviousSignals;

 private:
  void hideKana();
  void emptyBar();

 public:
 Missile(s32 x=0, s32 y=350)
   : m_nX(x), m_nY(y), m_nState(Ready), m_bLocked(false), m_nTargetY(0),
    m_nLoadCount(0), m_nBarRefreshing(0),
    m_bNeedLoad(false), m_reading(""), m_bPaused(false)
  {}
  HitBox getHitBox() const
  {
    return HitBox(getCenterX() - (WIDTH_MISSILE_SPRITE/4),
		  m_nY,
		  getCenterX() + (WIDTH_MISSILE_SPRITE/4),
		  m_nY + HEIGHT_MISSILE_SPRITE);
  }
  s32 getCenterX() const { return m_nX + (WIDTH_MISSILE_SPRITE/2); }
  void pause();
  void unpause();

  void fire(s32 x);
  void resetLocation();
  void setLoad() { m_bNeedLoad=true; }
  void setReload() { m_bNeedReload=true; }
  void lock() { m_bLocked = true; }
  void unlock() { m_bLocked = false; }
  void move();
  void startGoodCollision();
  void startBadCollision();
  void startLoad();
  void startReload();
  void setReading(Reading r);
  void showReading();
  bool atAnimationEnd() const { return (PA_DualGetSpriteAnimFrame(SPRITE_MISSILE) > 3); }
  bool atLoadAnimationEnd() const
  {
    return (PA_DualGetSpriteAnimFrame(SPRITE_BEGIN_MISSILE_BAR + 3)==4);
  }
  bool atReloadAnimationEnd() const
  {
    return (PA_DualGetSpriteAnimFrame(SPRITE_BEGIN_MISSILE_BAR + 3)==8);
  }
  bool collisionGood() const { return m_nState == GoodCollision; }

  bool unlocked() const { return !m_bLocked; }
  bool ready() const { return (m_nState == Ready); }
  bool inAir() const { return (m_nState == InAir); }
  bool loading() const { return (m_nState == Load); }
  bool exploding() const
  { 
    return (m_nState == GoodCollision) || (m_nState == BadCollision);
  }
  Reading reading() const { return m_reading; }

  void display() const
  {
    PA_OutputText(0,0,23, "MLock:%d MState:%d MRdSz=%d",
		  m_bLocked, m_nState, m_reading.size());
  }
  void addSignals(SignalSet &signals);
  void handleSignals(const SignalSet &signals);
  void setReading(std::string s);
};

typedef std::vector<Meteor> MeteorVec;


class ComboGraphic
{
 public:
  static const u8 NUM_DIGIT_SPRITES=3;

 private:
  s32 m_nX;
  s32 m_nY;
  s32 m_nRightOffset;
  u8 m_nBackgroundSprite;
  u8 m_nBeginDigitSprites;

 private:
  void setSpriteLocation(s32 x, s32 y);

 public:
  ComboGraphic(u8 background, u8 begin_digit);
  void show();
  void hide();
  void setCount(u16 c);
  void setLocation(s32 x, s32 y);

  static void initSprites(u8 background, u8 begin_digit);
};

class GameLoop
{
 private:
  static const u8 WIDTH_CURSOR_SPRITE=32;
  static const s32 DIST_STYLUS_CURSOR=100;
  static const s32 DIST_STYLUS_FIRE=32;
  static const bool DEBUG=false;

 private:
  enum State
  {
    Main,
    End
  };
  
  enum PauseAction { ToGame, ToMenu };

 private:
  MeteorVec m_vecMeteors;
  Missile m_missile;
  ComboGraphic m_combo;
  s16 m_nLifePoints;
  u8 m_nTimerID;
  s32 m_nCursorX;
  bool m_bRefreshData;
  std::vector<u8> m_vecChoosableMeteors;
  u8 m_nPreviousMeteor;
  u16 m_nCombo;
  bool m_bPaused;
  SignalSet m_setPreviousSignals;
  u8 m_bDSClosed;
  std::vector<u32> m_vecSignals;
  MenuInterface m_menuPause;

  u8 indexMissileCollision();
  void checkMissileCollision(SignalSet &signals);
  void moveMeteors();
  void moveCursor();
  void handleMissileInput();
  void resetMeteors();
  void resetMissile();
  void setCorrectIndex(u8 meteor_ind);
  void drawLifeBar();
  u16 missileRowMatch(u8 meteor_ind);

  bool gameplayCycle();
  bool pauseCycle();
  bool gameEndCycle();
  void addSignals(SignalSet &signals);
  void handleSignals(const SignalSet &signals);
  void cleanup();

 public:
  GameLoop();

  static void initSprites();

  void init(u8 num_meteors, u32 meteor_tpm);
  void run();

};

#endif
