//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com


#ifndef MENU_H
#define MENU_H

#include "menu_base.h"

namespace PlayerRecordDisplay
{
  const u8 ROW_GRADE=1;
  const u8 ROW_TOTAL=5;
  void show();
  void hide();
  void update();
  void initSprites();
};

enum MenuPage
{
  Gameplay,
  Data,
  Title,
  DeleteFile,
  DeleteConfirm,
  Game,
  LoadGame,
  ViewData,
  Guide,
  GuideGameplay,
  GuideControls,
  GuideAcknowledgements,
  GuideData
};

// Right trigger disabled temporarily until data menu completed.
class GameplayMenu
{
  enum Row { Speed, Meteors, New, Go };

 private:
  Trigger m_triggerLeft;
  Trigger m_triggerRight;
  MenuInterface m_interface;
  MenuPage m_nExit;

 private:
  bool loop();

 public:
  GameplayMenu();
  u32 getSpeed();
  u8 getMeteors();
  void run();
  MenuPage exit() const;
};

class DataMenu
{
  enum Row { ViewMastered, Delete };

 private:
  Trigger m_triggerLeft;
  Trigger m_triggerRight;
  MenuInterface m_interface;
  MenuPage m_nExit;

 private:
  bool loop();

 public:
  DataMenu();
  void run();
  MenuPage exit() const;
};

class DeleteConfirmMenu
{
  enum Row { DeleteYes, DeleteNo };

 private:
  MenuInterface m_interface;
  MenuPage m_nExit;

 private:
  bool loop();

 public:
  DeleteConfirmMenu();
  void run();
  MenuPage exit() const;
};

class TitleMenu
{
 private:
  MenuInterface m_interface;
  MenuPage m_nExit;

 private:
  bool loop();
  void init();

 public:
  TitleMenu();
  void run();
  MenuPage exit() const;
  u8 fileIndex() const;
};

class GuideMenu
{
 private:
  enum Row { GuideMenuGameplay,
	     GuideMenuData,
	     GuideMenuControls,
	     GuideMenuAcknowledgements,
	     GuideMenuBackToTitle };

 private:
  MenuInterface m_interface;
  MenuPage m_nExit;

 private:
  bool loop();

 public:
  GuideMenu();
  void run();
  MenuPage exit() const;
};

// A class for screens that are just text, with a single 'ok' button.
class TextPage
{
 private:
  std::vector<TextItem> m_vecTextItems;
  u8 m_nLine;
  MenuInterface m_interface;

 public:
  TextPage();
  void addTextItem(u8 indent, const std::string &s);
  void addBlankLines(u8 lines);
  void setBackText(const std::string &s);
  void run();
};


class DataViewer
{
 private:
  class Entry
  {
  private:
    u16 m_nUnicode;
    std::set<u16> m_setTopRows;
    std::vector< ReadingVec > m_vecPages;
  public:
    Entry(const std::set<u16> &tr,
	  const u8 &max_rows_per_page,
	  const u8 &max_kana_per_page);

    u8 pageCount() const;

    ReadingVec headerReadings() const;
    ReadingVec otherReadings(u8 page) const;

    u8 numRows(u8 page) const;
    u8 numKana(u8 page) const;
    u8 numHeaders() const;
    u16 unicode() const;
  };

 private:
  static const s32 KANJI_X=16;
  static const s32 KANA_X=64;
  static const s32 MAX_Y=304;
  static const u8 KANA_ROWS=19;
  static const u16 BAD_INDEX=-1;

 private:
  ButtonArray m_buttonsMenu;
  Button m_buttonLeft;
  Button m_buttonRight;
  Button m_buttonBack;
  std::vector<u16> m_vecRows;
  u16 m_nNumRows;
  u8 m_nKanaSpritesUsed;
  u8 m_nKanjiSpritesUsed;
  u16 m_nIndex;
  u16 m_nNumOnPage;
  u8 m_nPageWithinRow;

 private:

  Entry entry(u16 index) const;

  u16 beginUnicodeRange(u16 index) const;
  u16 endUnicodeRange(u16 index);

  u16 firstKana(u16 index);

  u16 beginKanaRange(u16 index);

  // what is the first kana of the current row?
  u16 previousKanaIndex();

  // what is the next different kana following the current row?
  u16 nextKanaIndex();

  ContentRow row(u16 index) const;
  u16 unicode(u16 index);

  // number of indices that can be put on a page starting at _index_.
  u8 pageLength(u16 index) const;

  // number of pages it will take to display _index_.
  u8 pageCount(u16 index) const;

  // Returns the index for the top of the page previous
  // to the page beginning with _index_.
  // Argument and return value both refer to m_vecRows.
  u16 previousPageTop(const u16 &index) const;

  // Returns true if it actually wrote.
  bool drawReading(u16 content_row_index, s32 x, s32 y);
  bool drawReading(const Reading &r, s32 x, s32 y);

  void drawEntry(const Entry &e, u8 page, u8 start_row);

  void clearPage();
  void drawPage();

  void updateKanaButtons();
  bool loop();

 public:
  DataViewer();
  void run();
  static void initSprites();
};

class MenuLoop
{
 private:
  GameplayMenu m_menuGameplay;
  DataMenu m_menuData;
  DeleteConfirmMenu m_menuDeleteConfirm;
  TitleMenu m_menuTitle;
  GuideMenu m_menuGuide;
  DataViewer m_viewerData;
  TextPage m_textGameplay;
  TextPage m_textControls;
  TextPage m_textAcknowledgements;
  TextPage m_textData;
  bool m_bFirstLoad;

 public:
  MenuLoop();
  void run();

  u32 getSpeed();
  u8 getMeteors();
  u8 getFileIndex() const;
};



#endif
