//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com


#include "menu_base.h"

//-------------------------------------------------------------------

HighlightBar::HighlightBar(u8 timer, u8 begin, u8 count)
  : m_nStartTapTime(0), m_bDown(false),
    m_nTimerID(timer), m_nBeginSprite(begin),
    m_nSpriteCount(count)
{}

//-------------------------------------------------------------------

void
HighlightBar::setTap()
{
  if (!m_bDown)
  {
    setFrame(1);
    m_bDown = true;
    m_nStartTapTime = Tick(m_nTimerID);
  }
}

//-------------------------------------------------------------------

void
HighlightBar::setY(s32 y)
{
  for (u8 i=0; i < m_nSpriteCount; ++i)
  {
    PA_DualSetSpriteY(m_nBeginSprite + i, y);
  }
}

//-------------------------------------------------------------------

void
HighlightBar::setFrame(u8 frame)
{
  for (u8 i=0; i < m_nSpriteCount; ++i)
  {
    PA_DualSetSpriteAnim(m_nBeginSprite + i, frame);
  }
}

//-------------------------------------------------------------------

bool
HighlightBar::clickEvent()
{
  if (m_bDown && m_nStartTapTime + PAUSE_TIME < Tick(m_nTimerID))
  {
    setFrame(0);
    m_bDown = false;
    return true;
  }
  else { return false; }
}

//-------------------------------------------------------------------

bool
HighlightBar::hasStylusTap() const
{
  if (Stylus.Newpress)
  {
    bool res(false);
    for (u8 i=0; i < m_nSpriteCount; ++i)
    {
      res = (res || PA_SpriteTouched(m_nBeginSprite + i));
    }
    return res;
  }
  else { return false; }
}

//-------------------------------------------------------------------

s32 HighlightBar::width() const { return 32 * m_nSpriteCount; }

//-------------------------------------------------------------------

s32 HighlightBar::height() const { return 16; }

//-------------------------------------------------------------------

void
HighlightBar::initSprites(u8 palette_id, u8 begin, u8 count,
			  const unsigned char* sprite, s32 x)
{
  for (u8 i=0; i < count; ++i)
  {
    PA_DualCreateSprite(begin + i,
			(void*)sprite,
			OBJ_SIZE_32X16,1, palette_id,
			x + (i*32), 384);
    PA_DualSetSpritePrio(begin + i, 2);
  }
}

//-------------------------------------------------------------------

u8 Trigger::s_nBeginLeftSprite;
u8 Trigger::s_nBeginRightSprite;

//-------------------------------------------------------------------

Trigger::Trigger(const std::string &s, bool left)
  : m_strContent(s),
    m_background(0, (left ? s_nBeginLeftSprite : s_nBeginRightSprite), 2),
    m_bLeft(left)
{}

//-------------------------------------------------------------------

void
Trigger::drawText(const std::string &s)
{
  if (m_bLeft)
  {
    PA_OutputText(0, 1, 1, const_cast<char*>(s.c_str()));
  }
  else
  {
    u8 x = 31 - s.size();
    PA_OutputText(0, x, 1, const_cast<char*>(s.c_str()));
  }
}

//-------------------------------------------------------------------

void
Trigger::show()
{
  drawText(m_strContent);
  m_background.setY(196);
}

//-------------------------------------------------------------------

void
Trigger::hide()
{
  drawText(std::string(m_strContent.size(), ' '));
  m_background.setY(384);
}

//-------------------------------------------------------------------

bool
Trigger::contains(s32 x, s32 y)
{ 
  return (x < m_background.width()
	  && y >= 192 
	  && y < 192 + m_background.height());
}

//-------------------------------------------------------------------

void
Trigger::tap() { m_background.setTap(); }

//-------------------------------------------------------------------

bool
Trigger::clickEvent() { return m_background.clickEvent(); }

//-------------------------------------------------------------------

void
Trigger::handleStandardInput()
{
  if ((m_bLeft && Pad.Newpress.L) || (!m_bLeft && Pad.Newpress.R)
      || m_background.hasStylusTap())
  {
    tap();
  }
}

//-------------------------------------------------------------------

void
Trigger::initLeftSprites(u8 palette_id, u8 begin, u8 count,
			 const unsigned char* sprite)
{
  HighlightBar::initSprites(palette_id, begin, count, sprite, 0);
  s_nBeginLeftSprite = begin;
}

//-------------------------------------------------------------------

void
Trigger::initRightSprites(u8 palette_id, u8 begin, u8 count,
			 const unsigned char* sprite)
{
  HighlightBar::initSprites(palette_id, begin, count, sprite, 192);
  s_nBeginRightSprite = begin;
}

//-------------------------------------------------------------------

MenuInterface::Style::Style()
  : v_align(Centered),
    margin_top(0),
    margin_bottom(24),
    label_option_space(2),
    max_vertical_space(24)
{}

//-------------------------------------------------------------------

MenuInterface::Style::Style(VerticalAlign va,
			    u8 mt,
			    u8 mb,
			    u8 los,
			    u8 mvs)
  : v_align(va),
    margin_top(mt),
    margin_bottom(mb),
    label_option_space(los),
    max_vertical_space(mvs)
{}

//-------------------------------------------------------------------

const std::string MenuInterface::MenuItem::OPTION_PREFIX="< ";
const std::string MenuInterface::MenuItem::OPTION_SUFFIX=" >";

u8 MenuInterface::s_nBeginHighlightBar;

//------------------------------------------------------------------

u8
MenuInterface::MenuItem::optionMarkerLength()
{
  return OPTION_PREFIX.size() + OPTION_SUFFIX.size();
}

//------------------------------------------------------------------

MenuInterface::MenuItem::MenuItem(const std::string &l,
				  const std::vector<std::string> &o)
  : m_strLabel(l), m_vecOptions(o), m_nIndex(0)
{}

//------------------------------------------------------------------

void
MenuInterface::MenuItem::left()
{
  if (!m_vecOptions.empty())
  {
    m_nIndex -= 1;
    if (m_nIndex < 0) { m_nIndex += m_vecOptions.size(); }
  }
}


//------------------------------------------------------------------

void
MenuInterface::MenuItem::right()
{
  if (!m_vecOptions.empty())
  {
    m_nIndex += 1;
    s8 size = m_vecOptions.size();
    if (m_nIndex >= size) { m_nIndex -= size; }
  }
}

//------------------------------------------------------------------

u8
MenuInterface::MenuItem::maxOptionLength() const
{
  u8 res=0;
  for (std::vector<std::string>::const_iterator iter=m_vecOptions.begin();
       iter != m_vecOptions.end();
       ++iter)
  {
    res = (res < iter->size() ? iter->size() : res);
  }
  return res;
}

//------------------------------------------------------------------

u8
MenuInterface::MenuItem::labelLength() const
{
  return m_strLabel.size();
}

//------------------------------------------------------------------

bool
MenuInterface::MenuItem::hasOptions() const
{
  return !m_vecOptions.empty();
}

//------------------------------------------------------------------

u8
MenuInterface::MenuItem::index() const
{
  return m_nIndex;
}

//------------------------------------------------------------------

void
MenuInterface::MenuItem::draw(const u8 &row,
			      const u8 &col1,
			      const u8 &col2,
			      const u8 &option_length)
{
  if (m_vecOptions.empty())
  {
    PA_OutputText(0, col1, row, const_cast<char*>(m_strLabel.c_str()));
  }
  else
  {
    std::string option = m_vecOptions[m_nIndex];
    u8 size_diff = option_length - option.size();
    for (u8 i=0; i < size_diff; ++i)
    {
      option = " " + option;
    }
    option = OPTION_PREFIX + option + OPTION_SUFFIX;
    PA_OutputText(0, col1, row, const_cast<char*>(m_strLabel.c_str()));
    PA_OutputText(0, col2, row, const_cast<char*>(option.c_str()));
  }
}

//------------------------------------------------------------------

MenuInterface::MenuInterface()
  : m_highlight(0, s_nBeginHighlightBar, 8),
    m_nCurrentIndex(0)
  , m_nMaxLabelLength(0)
  , m_nMaxOptionLength(0)
{}

//------------------------------------------------------------------

void MenuInterface::setStyle(const Style &s) { m_style = s; }

//------------------------------------------------------------------

void MenuInterface::clear() { m_vecItems.clear(); }

//------------------------------------------------------------------

u8
MenuInterface::addItem(const MenuItem &item)
{
  m_vecItems.push_back(item);
  if (item.hasOptions())
  {
    m_nMaxOptionLength = (m_nMaxOptionLength < item.maxOptionLength()
			  ? item.maxOptionLength()
			  : m_nMaxOptionLength);
  }
  m_nMaxLabelLength = (m_nMaxLabelLength < item.labelLength()
		       ? item.labelLength()
		       : m_nMaxLabelLength);

  updateHighlight();
  return m_vecItems.size() - 1;
}

//------------------------------------------------------------------

u8
MenuInterface::addTextItem(const std::string &text)
{
  return addSelectItem(text, std::vector<std::string>());
}

//------------------------------------------------------------------

u8
MenuInterface::addTextItem(u8 id, const std::string &text)
{
  return addSelectItem(id, text, std::vector<std::string>());
}

//------------------------------------------------------------------

u8
MenuInterface::addSelectItem(const std::string &text,
			     const std::vector<std::string> &options)
{
  u8 index = addItem(MenuItem(text, options));
  m_mapIDs[index] = index;
  return index;
}

//------------------------------------------------------------------

u8
MenuInterface::addSelectItem(u8 id,
			     const std::string &text,
			     const std::vector<std::string> &options)
{
  u8 index = addItem(MenuItem(text, options));
  m_mapIDs[id] = index;
  return index;
}

//------------------------------------------------------------------

u8
MenuInterface::verticalSpace() const
{
  if (m_vecItems.empty()) { return 0; }
  else
  {
    u8 items = m_vecItems.size();
    u8 space = m_style.margin_bottom
      - m_style.margin_top
      - m_vecItems.size();

    u8 res = (space / items);
    if (res > m_style.max_vertical_space)
    {
      res = m_style.max_vertical_space;
    }
    return res;
  }  
}

//------------------------------------------------------------------

u8
MenuInterface::row(u8 i) const
{
  u8 row_space = verticalSpace();
  u8 top_offset(0);
  u8 free_space = m_style.margin_bottom
    - m_style.margin_top 
    - (m_vecItems.size() * ( row_space + 1 ));

  if (m_style.v_align == Style::Centered) { top_offset = free_space / 2; }
  else if (m_style.v_align == Style::Bottom) { top_offset = free_space; }

  return m_style.margin_top + top_offset
    + (row_space >> 1) + i + (row_space * i);

}

//------------------------------------------------------------------

void
MenuInterface::drawText()
{
  u8 col1 = (32
	     - m_nMaxLabelLength
	     - m_nMaxOptionLength
	     - MenuItem::optionMarkerLength()) / 2;
  u8 col2 = col1 + m_nMaxLabelLength + m_style.label_option_space;
  for (u8 i=0; i < m_vecItems.size(); ++i)
  {
    m_vecItems[i].draw(row(i), col1, col2, m_nMaxOptionLength);
  }
}

//------------------------------------------------------------------

void
MenuInterface::eraseText()
{
  for (u8 i=0; i < m_vecItems.size(); ++i)
  {
    PA_OutputText(0, 0, row(i), "                                ");
  }
}

//------------------------------------------------------------------

void
MenuInterface::redraw()
{
  eraseText();
  drawText();
  updateHighlight();
}

//------------------------------------------------------------------

void
MenuInterface::right()
{
  m_vecItems[m_nCurrentIndex].right();
  redraw();
}

//------------------------------------------------------------------

void
MenuInterface::left()
{
  m_vecItems[m_nCurrentIndex].left();
  redraw();
}


//------------------------------------------------------------------

void
MenuInterface::up()
{
  if (m_nCurrentIndex > 0) { m_nCurrentIndex -= 1; }
  redraw();
}

//------------------------------------------------------------------

void
MenuInterface::down()
{
  if (m_nCurrentIndex < m_vecItems.size() - 1) { m_nCurrentIndex += 1; }
  redraw();
}

//------------------------------------------------------------------

void
MenuInterface::a()
{
  if (!m_vecItems[m_nCurrentIndex].hasOptions())
  {
    m_highlight.setTap();
    redraw();
  }
}

//------------------------------------------------------------------

void
MenuInterface::tap(s32 x, s32 y)
{
  for (u8 i=0;
       i < m_vecItems.size();
       ++i)
  {
    s32 current_top = (8 * row(i)) - 4;
    if (y >= current_top && y < current_top + 16)
    {
      m_highlight.setTap();
      m_nCurrentIndex = i;
      right();
      redraw();
    }
  }
}

//------------------------------------------------------------------

bool MenuInterface::clickEvent() { return m_highlight.clickEvent(); }

//------------------------------------------------------------------

u8
MenuInterface::index(u8 id) const
{
  std::map<u8, u8>::const_iterator iter = m_mapIDs.find(id);
  if (iter != m_mapIDs.end())
  {
    return m_vecItems[iter->second].index();
  }
  else { return -1; }
}

//------------------------------------------------------------------

u8
MenuInterface::currentIndex() const
{
  u8 res=-1;
  for (std::map<u8, u8>::const_iterator iter=m_mapIDs.begin();
       iter != m_mapIDs.end() && res == (u8)-1; ++iter)
  {
    if (iter->second == m_nCurrentIndex) { res = iter->first; }
  }
  return res;
}

//------------------------------------------------------------------

void
MenuInterface::hide()
{
  eraseText();
  m_highlight.setY(384);
}

//------------------------------------------------------------------

void
MenuInterface::show()
{
  redraw();
}

//------------------------------------------------------------------

void
MenuInterface::handleStandardInput()
{
  if (Pad.Newpress.Right) { right(); }
  if (Pad.Newpress.Left) { left(); }
  if (Pad.Newpress.Up) { up(); }
  if (Pad.Newpress.Down) { down(); }
  if (Pad.Newpress.A) { a(); }
  if (Stylus.Newpress) { tap(Stylus.X, Stylus.Y); }
}

//------------------------------------------------------------------

void
MenuInterface::updateHighlight()
{
  m_highlight.setY((8 * row(m_nCurrentIndex)) - 4 + 192);
}

//------------------------------------------------------------------

void
MenuInterface::initSprites(u8 palette_id,
			   u8 begin_id,
			   const unsigned char* sprite)
{
  HighlightBar::initSprites(palette_id, begin_id, 8, sprite, 0);
  s_nBeginHighlightBar = begin_id;
}

//------------------------------------------------------------------

bool inputKeyNewpress(InputKey i)
{
  switch (i)
  {
  case KeyUp: return Pad.Newpress.Up; break;
  case KeyDown: return Pad.Newpress.Down; break;
  case KeyLeft: return Pad.Newpress.Left; break;
  case KeyRight: return Pad.Newpress.Right; break;
  case KeyTriggerLeft: return (bool)Pad.Newpress.L; break;
  case KeyTriggerRight: return (bool)Pad.Newpress.R; break;
  case KeyA: return Pad.Newpress.A; break;
  case KeyB: return Pad.Newpress.B; break;
  default: return false;
  }
}

//------------------------------------------------------------------

Button::Button(u8 b, s32 bw, s32 bh, u8 c, s32 x, s32 y, InputKey i,
	       bool fp, s32 xoff, s32 yoff)
  : m_nBackgroundSprite(b),
    m_nBackgroundWidth(bw),
    m_nBackgroundHeight(bh),
    m_nContentSprite(c),
    m_nX(x),
    m_nY(y),
    m_nContentXOffset(xoff),
    m_nContentYOffset(yoff),
    m_nInputKey(i),
    m_bFlashPermanent(fp),
    m_bDown(false),
    m_nStartTapTime(0),
    m_nTimerID(0),
    m_bActive(true)
{}

//------------------------------------------------------------------

void
Button::setContentSprite(u8 c, u8 frame, s32 xoff, s32 yoff)
{
  // hide the current sprite?  probably.
  // unset highlight settings?
  if (m_nContentSprite != c)
  {
    PA_DualSetSpriteXY(m_nContentSprite, m_nX, 384);
  }
  m_nContentSprite = c;
  m_nContentXOffset = xoff;
  m_nContentYOffset = yoff;
  PA_DualSetSpriteAnim(m_nContentSprite, frame);
  PA_DualSetSpriteXY(m_nContentSprite, m_nX + xoff, m_nY + yoff);
}

//------------------------------------------------------------------

bool
Button::hasStylusTap() const
{
  if (Stylus.Newpress)
  {
    return (PA_SpriteTouched(m_nBackgroundSprite) > 0);
  }
  else { return false; }
}

//------------------------------------------------------------------

bool Button::hasKeyTap() const { return inputKeyNewpress(m_nInputKey); }

//------------------------------------------------------------------

bool
Button::handleStandardInput()
{
  bool res(false);
  if (hasStylusTap() || hasKeyTap())
  {
    flashOn();
    res = true;
  }
  return res;
}

//------------------------------------------------------------------

bool
Button::busy() const
{
  if (m_bFlashPermanent) { return m_bDown; }
  else 
  {
    return (m_bDown && (m_nStartTapTime + PAUSE_TIME > Tick(m_nTimerID))); 
  }
}

//------------------------------------------------------------------

void
Button::flashOn()
{
  if (!m_bDown)
  {
    PA_DualSetSpriteAnim(m_nBackgroundSprite, 1);
    m_bDown = true;
    m_nStartTapTime = Tick(m_nTimerID);
  }
}

//------------------------------------------------------------------

void
Button::flashOff()
{
  PA_DualSetSpriteAnim(m_nBackgroundSprite, 0);
  m_bDown = false;
}

//------------------------------------------------------------------

bool
Button::checkClick()
{
  if (m_bFlashPermanent) { return m_bDown; }
  else if (m_bDown && m_nStartTapTime + PAUSE_TIME < Tick(m_nTimerID))
  {
    PA_DualSetSpriteAnim(m_nBackgroundSprite, 0);
    m_bDown = false;
    return true;
  }
  else { return false; }
}

//------------------------------------------------------------------

void
Button::show()
{
  m_bActive = true;
  PA_DualSetSpriteXY(m_nBackgroundSprite, m_nX, m_nY);
  PA_DualSetSpriteXY(m_nContentSprite,
		     m_nX + m_nContentXOffset,
		     m_nY + m_nContentYOffset);
}

//------------------------------------------------------------------

void
Button::hide()
{
  m_bActive = false;
  PA_DualSetSpriteXY(m_nBackgroundSprite, m_nX, 384);
  PA_DualSetSpriteXY(m_nContentSprite, m_nX, 384);
}

//------------------------------------------------------------------

ButtonArray::ButtonArray()
  : m_nCurrentIndex(0)
{}

//------------------------------------------------------------------

void ButtonArray::addButton(Button b) { m_vecButtons.push_back(b); }

//------------------------------------------------------------------

void
ButtonArray::previous()
{
  if (m_nCurrentIndex > 0)
  {
    m_vecButtons[m_nCurrentIndex].flashOff();
    m_vecButtons[m_nCurrentIndex-1].flashOn();
    m_nCurrentIndex -= 1;
  }
}

//------------------------------------------------------------------

void
ButtonArray::next()
{
  if (m_nCurrentIndex < m_vecButtons.size() - 1)
  {
    m_vecButtons[m_nCurrentIndex].flashOff();
    m_vecButtons[m_nCurrentIndex+1].flashOn();
    m_nCurrentIndex += 1;
  }
}

//------------------------------------------------------------------

bool
ButtonArray::handleStylusInput()
{
  u8 old_index(m_nCurrentIndex);
  u8 new_index(m_nCurrentIndex);
  bool res(false);
  for (u8 i=0; i < m_vecButtons.size(); ++i)
  {
    if (m_vecButtons[i].hasStylusTap())
    {
      new_index=i;
      res = true;
    }
  }
  if (old_index != new_index)
  {
    m_vecButtons[old_index].flashOff();
    m_vecButtons[new_index].flashOn();
    m_nCurrentIndex = new_index;
  }
  return res;
}

//------------------------------------------------------------------

bool
ButtonArray::handleTriggerInput()
{
  if (inputKeyNewpress(KeyTriggerLeft) && Pad.Newpress.L)
  {
    previous();
    return true;
  }
  else if (inputKeyNewpress(KeyTriggerRight))
  {
    next();
    return true;
  }
  else { return false; }
}

//------------------------------------------------------------------

void
ButtonArray::show()
{
  for (std::vector<Button>::iterator iter=m_vecButtons.begin();
       iter != m_vecButtons.end();
       ++iter)
  {
    iter->show();
  }
  if (!m_vecButtons.empty()) { m_vecButtons[m_nCurrentIndex].flashOn(); }
}

//------------------------------------------------------------------

void
ButtonArray::hide()
{
  for (std::vector<Button>::iterator iter=m_vecButtons.begin();
       iter != m_vecButtons.end();
       ++iter)
  {
    iter->hide();
  }
}

//------------------------------------------------------------------

bool
ButtonArray::checkClick()
{
  bool res(false);
  for (std::vector<Button>::iterator iter=m_vecButtons.begin();
       iter != m_vecButtons.end();
       ++iter)
  {
    res = res | iter->checkClick();
  }
  return res;
}

//------------------------------------------------------------------

u8 ButtonArray::index() const { return m_nCurrentIndex; }

//------------------------------------------------------------------

TextItem::TextItem(u8 row, u8 indent,
		   const std::string &full_text)
  : m_nBeginRow(row),
    m_nIndent(indent)
{
  unsigned int line_begin = full_text.find_first_not_of(" ");
  while (line_begin != std::string::npos)
  {
    std::string s = full_text.substr(line_begin,
				     CHARS_PER_LINE - m_nIndent);
    unsigned int line_end = s.find_last_of(' ');
    if (line_end != std::string::npos
	&& s.size() == CHARS_PER_LINE - m_nIndent)
    {
      s = s.substr(0, line_end);
    }
    m_vecText.push_back(s);
    line_begin += s.size();
    if (line_begin >= full_text.size()) { line_begin = std::string::npos; }
    else
    {
      line_begin = full_text.find_first_not_of(" ", line_begin);
    }
  }
  
}

//-------------------------------------------------------------------

u8 TextItem::numRows() { return m_vecText.size(); }

//-------------------------------------------------------------------

void
TextItem::show()
{
  for (u8 i=0; i < m_vecText.size(); ++i)
  {
    u8 screen(0);
    u8 line = m_nBeginRow+i;
    if (line < 24) { screen = 1; }
    else { line -= 24;}

    PA_OutputText(screen, m_nIndent, line,
		  const_cast<char *>(m_vecText[i].c_str()));
  }
}

//-------------------------------------------------------------------

void
TextItem::hide()
{
  for (u8 i=0; i < m_vecText.size(); ++i)
  {
    u8 screen(0);
    u8 line = m_nBeginRow+i;
    if (line < 24) { screen = 1; }
    else { line -= 24; }

    std::string s(m_vecText[i].size(), ' ');
    PA_OutputText(screen, m_nIndent, line,  const_cast<char *>(s.c_str()));
  }
}
