//    Kanji Strike - game to learn kanji readings
//    Copyright (C) 2008-2009  Kimberly Dorn

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//    Contact email: kkdorn@gmail.com


#ifndef MENU_BASE_H
#define MENU_BASE_H

#include <PA9.h>
#include <map>
#include <string>
#include <vector>

class HighlightBar
{
  static const u32 PAUSE_TIME=100;

 private:
  u32 m_nStartTapTime;
  bool m_bDown;
  u8 m_nTimerID;
  u8 m_nBeginSprite;
  u8 m_nSpriteCount;

 private:
  void setFrame(u8 frame);

 public:
  HighlightBar(u8 timer, u8 begin, u8 count);
  void setTap();
  void setY(s32 y);
  bool clickEvent();
  bool hasStylusTap() const;
  s32 width() const;
  s32 height() const;
  static void initSprites(u8 palette_id, u8 begin, u8 count,
			  const unsigned char* sprite, s32 x);
};

class Trigger
{
 private:
  std::string m_strContent;
  HighlightBar m_background;
  bool m_bLeft;

 private:
  static u8 s_nBeginLeftSprite;
  static u8 s_nBeginRightSprite;

 private:
  void drawText(const std::string &s);

 public:
  Trigger(const std::string &s, bool left);
  void show();
  void hide();
  bool contains(s32 x, s32 y);
  void tap();
  bool clickEvent();
  void handleStandardInput();

  static void initLeftSprites(u8 palette_id, u8 begin, u8 count,
			      const unsigned char* sprite);
  static void initRightSprites(u8 palette_id, u8 begin, u8 count,
			       const unsigned char* sprite);
};

class MenuInterface
{
 public:
  class Style
  {
  public:
    enum VerticalAlign { Centered, Top, Bottom };

    // different naming format for public member data.
  public:
    VerticalAlign v_align;
    u8 margin_top;
    u8 margin_bottom;
    u8 label_option_space;
    u8 max_vertical_space;
    Style();
    Style(VerticalAlign va, u8 mt, u8 mb, u8 los, u8 mvs);
  };

 private:
  class MenuItem
  {
  private:
    static const std::string OPTION_PREFIX;
    static const std::string OPTION_SUFFIX;

  private:
    std::string m_strLabel;
    std::vector<std::string> m_vecOptions;
    s8 m_nIndex;

  public:
    static u8 optionMarkerLength();

    MenuItem(const std::string &l, const std::vector<std::string> &o);
    void left();
    void right();
    void draw() const;
    u8 maxOptionLength() const;
    u8 labelLength() const;
    u8 index() const;
    bool hasOptions() const;
    void draw(const u8 &row, const u8 &col1, const u8 &col2,
	      const u8 &option_length);
  };

 private:
  static const u8 NUM_HIGHLIGHT_BAR_SPRITES=8;
  static u8 s_nBeginHighlightBar;

 private:
  std::vector<MenuItem> m_vecItems;
  std::map<u8, u8> m_mapIDs;
  HighlightBar m_highlight;
  u8 m_nCurrentIndex;
  u8 m_nMaxLabelLength;
  u8 m_nMaxOptionLength;
  Style m_style;

 private:
  u8 verticalSpace() const;
  u8 row(u8 i) const;
  u8 addItem(const MenuItem &item);
  void updateHighlight();
  void drawText();
  void eraseText();
  void redraw();

 public:
  MenuInterface();
  void setStyle(const Style &s);
  void clear();
  u8 addTextItem(u8 id, const std::string &text);
  u8 addTextItem(const std::string &text);
  u8 addSelectItem(u8 id, const std::string &text,
		   const std::vector<std::string> &options);
  u8 addSelectItem(const std::string &text,
		   const std::vector<std::string> &options);

  void up();
  void down();
  void left();
  void right();
  void a();
  void tap(s32 x, s32 y);
  bool clickEvent();
  u8 index(u8 id) const;
  u8 currentIndex() const;

  void hide();
  void show();

  void handleStandardInput();

  static void initSprites(u8 palette_id,
			  u8 begin_id,
			  const unsigned char* sprite);
};

enum InputKey
{
  KeyNone,
  KeyUp,
  KeyDown,
  KeyLeft,
  KeyRight,
  KeyTriggerLeft,
  KeyTriggerRight,
  KeyA,
  KeyB
};

bool inputKeyNewpress(InputKey i);

class Button
{
 private:
  static const u32 PAUSE_TIME=100;

 private:
  u8 m_nBackgroundSprite;
  s32 m_nBackgroundWidth;
  s32 m_nBackgroundHeight;
  u8 m_nContentSprite;
  s32 m_nX;
  s32 m_nY;
  s32 m_nContentXOffset;
  s32 m_nContentYOffset;
  InputKey m_nInputKey;
  bool m_bFlashPermanent;
  bool m_bDown;
  u32 m_nStartTapTime;
  u8 m_nTimerID;
  bool m_bActive;

 public:
  Button(u8 b, s32 bw, s32 bh, u8 c, s32 x, s32 y, InputKey i, bool fp,
	 s32 xoff, s32 yoff);
  void setContentSprite(u8 c, u8 frame, s32 xoff, s32 yoff);
  bool hasStylusTap() const;
  bool hasKeyTap() const;
  bool handleStandardInput();
  bool busy() const;
  void flashOn();

  // necessary?
  void flashOff();

  // In permanent flash mode, check click returns true
  // as long as the button is selected.
  // In temporary flash mode, check click returns true once
  //  after a button has been tapped.
  bool checkClick();

  void show();
  void hide();
};

class ButtonArray
{
 private:
  std::vector<Button> m_vecButtons;
  u8 m_nCurrentIndex;

 public:
  ButtonArray();
  void addButton(Button b);
  void previous();
  void next();

  // Return true if they handled any input.
  bool handleStylusInput();
  bool handleTriggerInput();

  void show();
  void hide();
  bool checkClick();
  u8 index() const;
};


class TextItem
{
 private:
  static const u8 CHARS_PER_LINE=32;

 private:
  u8 m_nBeginRow;
  u8 m_nIndent;
  std::vector<std::string> m_vecText;

 public:
  TextItem(u8 row, u8 indent,
	   const std::string &full_text);
  u8 numRows();
  void show();
  void hide();
};


#endif
